"""Convert a 4-column CSV metadata file to a YAML file."""

import argparse
import csv
import logging

import yaml

# Set up logging
logger = logging.getLogger("joint_pca")


def csv_to_yaml(arguments: argparse.Namespace):
    """Convert a CSV file with biosample, assay, experiment, accession columns to a YAML file.

    Args:
    ----
        arguments (argparse.Namespace): Arguments from the command line.

    """
    csv_file = arguments.csv_file  # Path to the input CSV file.
    yaml_file = arguments.yaml_file  # Path to the output YAML file.

    metadata = []
    with open(csv_file) as f:
        reader = csv.reader(f)
        for row in reader:
            biosample, assay, experiment, accession = row
            metadata.append({"biosample": biosample, "assay": assay, "experiment": experiment, "accession": accession})

    with open(yaml_file, "w") as f:
        yaml.dump(metadata, f, default_flow_style=False)


def configure_trackcsv2yaml_parser(parser):
    """Configure the parser for the trackcsv2yaml script."""
    parser.add_argument("csv_file", help="Path to the input CSV file.")
    parser.add_argument("yaml_file", help="Path to the output YAML file.")
    parser.set_defaults(func=csv_to_yaml)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Convert a 4-column CSV metadata file to a YAML file.")
    configure_trackcsv2yaml_parser(parser)
    args = parser.parse_args()
    csv_to_yaml(args.csv_file, args.yaml_file)
    logger.info(f"Metadata successfully written to {args.yaml_file}")



import os.path
import uuid
from glob import glob

import numpy as np
import pandas as pd
import pytest

from jointly_hic.core.config import JointlyConfiguration
from jointly_hic.core.decomposer import JointlyDecomposer
from jointly_hic.core.post_processor import PostProcessor
from jointly_hic.core.trajectory_analyzer import TrajectoryAnalyzer

SKIP_END2END = False
DATA_FILES = [
    "data/DE-FA-DSG-DdeI-DpnII-P2.hg38.mapq_30.1000.mcool",
    "data/ESC-FA-DSG-DdeI-DpnII-P2.hg38.mapq_30.1000.mcool",
    "data/HB-FA-DSG-DdeI-DpnII-P2.hg38.mapq_30.1000.mcool",
]
EXCLUSION_LIST = "data/blacklist.bed"


@pytest.fixture
def configuration():
    """Fixture for JointPCAConfig."""
    return JointlyConfiguration(
        mcools=DATA_FILES,
        output=f"{uuid.uuid4()!s}",
        resolution=250000,
        assembly="hg38",
        components=32,
        chrom_limit=4,
        method="PCA",
        percentile_top=99.5,
        percentile_bottom=1,
        batch_size=20000,
        log_level="DEBUG",
        exclusion_list=EXCLUSION_LIST,
    )


def check_data_files():
    """Check for data/ files and return True/False."""
    if SKIP_END2END:
        return False
    return all(os.path.exists(data_file) for data_file in DATA_FILES)


@pytest.mark.parametrize("method", ["PCA", "NMF", "SVD"])
@pytest.mark.skipif(not check_data_files(), reason="Data files not found")
def test_end_to_end_decomposition(configuration, method):
    """Test the JointPCADecomposer complete pipeline."""
    try:
        configuration.method = method
        decomposer = JointlyDecomposer(configuration)
        output_embeddings, post_config, trajectory_config = decomposer.run()

        # Make sure output files exist
        assert os.path.exists(post_config.parquet_file)

        # Load embeddings from parquet and check on them
        embeddings = pd.read_parquet(post_config.parquet_file)
        assert np.all(
            embeddings.columns
            == ["chrom", "start", "end", "weight", "good_bin", "filename"] + [f"{method}{i + 1}" for i in range(32)]
        )
        assert embeddings.shape == (10560, 38)
        assert embeddings.dropna().shape == (9498, 38)

        # Run post-processing
        post_processor = PostProcessor(post_config)
        post_processor.run()

        # Check output files
        assert os.path.exists(f"{post_config.output}_post_processed.csv.gz")
        assert os.path.exists(f"{post_config.output}_post_processed.pq")
        assert os.path.exists(f"{post_config.output}_scores.png")
        assert os.path.exists(f"{post_config.output}_scores_clustered.png")
        assert os.path.exists(f"{post_config.output}_scores_filenames.png")

        for n_neighbors in post_config.umap_neighbours:
            assert os.path.exists(f"{post_config.output}_umap-n{n_neighbors}_clustered.png")
            assert os.path.exists(f"{post_config.output}_umap-n{n_neighbors}_filenames.png")

        # Run TrajectoryAnalyzer
        trajectory_analyzer = TrajectoryAnalyzer(trajectory_config)
        trajectory_analyzer.run()

        # Check output files
        assert os.path.exists(f"{trajectory_config.output}_trajectories.csv.gz")
        assert os.path.exists(f"{trajectory_config.output}_trajectories.pq")

    except Exception as e:
        raise e

    finally:
        # Remove  output files
        filelist = glob(configuration.output + "*")

        # Remove mv5 and npy files
        for f in DATA_FILES:
            filelist.extend(glob(f.replace(".mcool", "*mv5")))
            filelist.extend(glob(f.replace(".mcool", "*npy")))

        # Remove  output files
        for f in filelist:
            os.remove(f)



import bioframe
import numpy as np
import pandas as pd
import pytest
from sklearn.decomposition import IncrementalPCA, MiniBatchNMF

from jointly_hic.core.config import JointlyConfiguration
from jointly_hic.core.decomposer import JointlyDecomposer
from jointly_hic.methods.sparse_incremental_svd import SparseIncrementalSVD


@pytest.fixture
def configuration():
    """Fixture for JointPCAConfig."""
    return JointlyConfiguration(
        mcools=["test1.mcool", "test2.mcool"],
        output="test",
        resolution=10000,
        assembly="hg38",
        components=32,
        chrom_limit=23,
        method="PCA",
        percentile_top=99.5,
        percentile_bottom=1,
        batch_size=20000,
        log_level="INFO",
    )


@pytest.fixture
def joint_pca_decomposer(mocker, configuration):
    """Fixture for JointPCADecomposer."""
    mocker.patch(
        "jointly_hic.core.decomposer.JointlyDecomposer.partition",
        return_value=[0, 1, 2, 3, 4, 5, 6, 7],
    )
    return JointlyDecomposer(configuration)


def test_joint_pca_decomposer(mocker, configuration):
    """Test JointPCADecomposer init."""
    # Mock JointPCADecomposer.get_partition
    get_partition_patch = mocker.patch(
        "jointly_hic.core.decomposer.JointlyDecomposer.partition",
        return_value=[0, 1, 2, 3, 4, 5, 6, 7],
    )
    joint_pca_decomposer = JointlyDecomposer(configuration)

    # Test init with PCA model
    assert joint_pca_decomposer.configuration == configuration
    assert np.all(joint_pca_decomposer.chromosome_sizes == bioframe.fetch_chromsizes("hg38")[0:23])
    assert isinstance(joint_pca_decomposer.model, IncrementalPCA)
    assert joint_pca_decomposer.partition == get_partition_patch


@pytest.mark.parametrize(
    "method, model",
    [
        ("PCA", IncrementalPCA),
        ("NMF", MiniBatchNMF),
        ("SVD", SparseIncrementalSVD),
    ],
)
def test_joint_nmf_decomposer(mocker, configuration, method, model):
    """Test JointPCADecomposer init when method is NMF."""
    configuration.method = method
    joint_pca_decomposer = JointlyDecomposer(configuration)
    assert isinstance(joint_pca_decomposer.model, model)


def test_run(mocker, joint_pca_decomposer, configuration):
    """Test the JointPCADecomposer.run."""
    joint_pca_decomposer = JointlyDecomposer(configuration)
    joint_pca_decomposer.save_model = mocker.Mock()
    joint_pca_decomposer.decompose_cooler_file = mocker.Mock()
    joint_pca_decomposer.set_union_bad_bins = mocker.MagicMock()
    joint_pca_decomposer.compute_output_embeddings = mocker.Mock()
    joint_pca_decomposer.run()
    joint_pca_decomposer.save_model.assert_called_once()
    joint_pca_decomposer.decompose_cooler_file.assert_has_calls(
        [
            mocker.call(
                filename="test1.mcool",
            ),
            mocker.call(
                filename="test2.mcool",
            ),
        ]
    )
    joint_pca_decomposer.set_union_bad_bins.assert_called_once()


def test_partition(mocker, configuration):
    # Test when _partition is set
    joint_pca_decomposer = JointlyDecomposer(configuration)
    joint_pca_decomposer._partition = np.ndarray([0, 1, 2, 3, 4, 5, 6, 7])
    assert np.all(joint_pca_decomposer.partition == np.ndarray([0, 1, 2, 3, 4, 5, 6, 7]))

    # Test when _partition is not set
    joint_pca_decomposer._partition = None
    # Patch Cooler
    cooler_patch = mocker.patch("jointly_hic.core.decomposer.Cooler")
    cooler_patch.return_value.offset.return_value = 10

    partition = joint_pca_decomposer.partition
    for x in partition:
        assert x == 10

    assert cooler_patch.call_count == 2


def test_get_chromosome_sizes(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.get_chromosome_sizes."""
    mocker.patch(
        "jointly_hic.core.decomposer.bioframe.fetch_chromsizes",
        return_value=pd.Series([1, 2, 3, 4, 5, 6]),
    )
    assert np.all(joint_pca_decomposer.get_chromosome_sizes() == pd.Series([1, 2, 3, 4, 5, 6]))


def test_set_union_bad_bins(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.take_union_bad_bins."""
    # Mock JointPCADecomposer.preprocess_matrix
    preprocess_matrix_patch = mocker.patch("jointly_hic.core.decomposer.JointlyDecomposer.preprocess_matrix")
    # Mock JointPCADecomposer.preprocess_matrix to return a dataframe with shape (50, 50) and a union_bad_bins
    preprocess_matrix_patch.return_value = (
        pd.DataFrame(np.random.random((50, 50))),
        pd.DataFrame(np.array([False, True] * 25)),
    )
    # Mock Cooler.cooler
    mocker.patch("jointly_hic.core.decomposer.Cooler")

    # Test take_union_bad_bins
    joint_pca_decomposer.set_union_bad_bins()
    # Assert that preprocess_matrix_patch was called with test1.mcool and test2.mcool
    preprocess_matrix_patch.assert_has_calls(
        [
            mocker.call("test1.mcool", ignore_cache=True),
            mocker.call("test2.mcool", ignore_cache=True),
        ]
    )


def test_preprocess_matrix(mocker, joint_pca_decomposer):
    # Patch numpy.save
    numpy_save_patch = mocker.patch("jointly_hic.core.decomposer.np.save")
    cooler_patch = mocker.patch("jointly_hic.core.decomposer.Cooler")
    normalized_affinity_matrix_patch = mocker.patch("jointly_hic.core.decomposer.normalized_affinity_matrix_from_trans")
    normalized_affinity_matrix_patch.return_value = (mocker.MagicMock(), mocker.MagicMock())
    logical_not_patch = mocker.patch("jointly_hic.core.decomposer.np.logical_not")

    matrix = joint_pca_decomposer.preprocess_matrix("test.mcool")
    cooler_patch.assert_called_once_with(f"test.mcool::/resolutions/{joint_pca_decomposer.configuration.resolution}")
    normalized_affinity_matrix_patch.assert_called_once_with(
        cooler_patch.return_value.matrix.return_value.__getitem__.return_value,
        joint_pca_decomposer.partition.__sub__(),
        joint_pca_decomposer.configuration.percentile_top,
        joint_pca_decomposer.configuration.percentile_bottom,
    )
    logical_not_patch.assert_called_once()
    numpy_save_patch.assert_called_once()


def test_minibatch_fit(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.minibatch_fit."""
    # Mock IncrementalPCA.partial_fit
    partial_fit_patch = mocker.patch("jointly_hic.core.decomposer.IncrementalPCA.partial_fit")
    matrix = np.random.random((100, 100))

    # Test minibatch_fit
    joint_pca_decomposer.minibatch_fit(matrix)
    partial_fit_patch.assert_called_once()


def test_decompose_cooler_file(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.decompose_cooler_file."""
    # Mock JointPCADecomposer.minibatch_fit
    minibatch_fit_patch = mocker.patch("jointly_hic.core.decomposer.JointlyDecomposer.minibatch_fit")
    # Mock JointPCADecomposer.preprocess_matrix
    preprocess_matrix_patch = mocker.patch("jointly_hic.core.decomposer.JointlyDecomposer.preprocess_matrix")

    joint_pca_decomposer.decompose_cooler_file("test.mcool")
    minibatch_fit_patch.assert_called_once()
    preprocess_matrix_patch.assert_called_once()


def test_save_model(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.save_model."""
    joint_pca_decomposer.model = mocker.Mock()
    # Mock gzip.open
    gzip_open_patch = mocker.patch("gzip.open")
    # Mock pickle.dump
    pickle_dump_patch = mocker.patch("pickle.dump")

    # Test save_model with gzip
    joint_pca_decomposer.save_model("testout.pkl.gz")
    gzip_open_patch.assert_called_once_with("testout.pkl.gz", "wb")
    pickle_dump_patch.assert_called_once()


def test_compute_output_embeddings(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.compute_output_embeddings."""
    joint_pca_decomposer.compute_output_embeddings_single_file = mocker.Mock(
        return_value=pd.DataFrame({"test": [1, 2, 3]})
    )
    embeddings = joint_pca_decomposer.compute_output_embeddings()
    joint_pca_decomposer.compute_output_embeddings_single_file.assert_has_calls(
        [
            mocker.call("test1.mcool"),
            mocker.call("test2.mcool"),
        ]
    )
    assert np.all(embeddings == pd.concat([pd.DataFrame({"test": [1, 2, 3]}), pd.DataFrame({"test": [1, 2, 3]})]))


def test_compute_output_embeddings_single_file(mocker, joint_pca_decomposer):
    """Test JointPCADecomposer.compute_output_embeddings_single_file."""
    # Mock JointPCADecomposer.preprocess_matrix
    preprocess_matrix_patch = mocker.patch("jointly_hic.core.decomposer.JointlyDecomposer.preprocess_matrix")
    # Mock IncrementalPCA.transform and return array with shape (50, 32)
    transform_patch = mocker.patch(
        "jointly_hic.core.decomposer.IncrementalPCA.transform",
        return_value=np.random.random((25, 32)),
    )
    # Patch make_multivec
    mocker.patch("jointly_hic.core.decomposer.make_multivec")
    # Add partitions
    joint_pca_decomposer.partition = np.array([0, 25, 40, 50])
    # Mock JointPCADecomposer.union_bad_bins of length 50
    joint_pca_decomposer.union_bad_bins = np.array([False, True] * 25)

    # Mock Cooler to return bins with length 50
    bins = pd.DataFrame({"chrom": ["chr1"] * 50, "start": np.arange(50), "end": np.arange(1, 51)})
    cooler_patch = mocker.patch("jointly_hic.core.decomposer.Cooler")
    # Return bins for Cooler.bins()[lo:hi]
    cooler_patch.return_value.bins.return_value.__getitem__.return_value = bins

    # Test compute_output_embeddings_single_file
    embeddings = joint_pca_decomposer.compute_output_embeddings_single_file("test.mcool")

    # Assertions
    preprocess_matrix_patch.assert_called_once_with("test.mcool")
    transform_patch.assert_called_once_with(preprocess_matrix_patch.return_value)
    assert embeddings.shape == (50, 37)
    assert embeddings.columns.tolist() == [
        "chrom",
        "start",
        "end",
        "good_bin",
        "filename",
    ] + [f"PCA{i + 1}" for i in range(32)]
    assert embeddings.good_bin.tolist() == [True, False] * 25
    assert embeddings.start.tolist() == np.arange(50).tolist()
    assert embeddings.end.tolist() == np.arange(1, 51).tolist()
    assert embeddings.chrom.tolist() == ["chr1"] * 50

    # Loop over embeddings, every other row matches a row in transorm_patch.return_value
    for i in range(25):
        assert np.all(embeddings.iloc[2 * i, 5:] == transform_patch.return_value[i])



from os import remove
from os.path import isfile
from uuid import uuid4

import h5py
import numpy as np
import pandas as pd

from jointly_hic.methods.multivec import nansum_agg, nanmean_agg, logsumexp_agg, make_multivec


def test_nansum_agg():
    # Create 3D array of test data
    data = np.array([[[1, np.nan], [3, 4], [5, 6]], [[7, np.nan], [9, 10], [11, 12]]])
    result = nansum_agg(data)
    assert np.all(result == np.array([[8, 16, 14], [12, 0, 18]]))


def test_nanmean_agg():
    # Create 3D array of test data
    data = np.array([[[1, np.nan], [3, 4], [5, 6]], [[7, np.nan], [9, 10], [11, 12]]])
    result = nanmean_agg(data)
    assert np.allclose(result, np.array([[4, 8, 7], [6, np.nan, 9]]), equal_nan=True)


def test_logsumexp_agg():
    # Create 3D array of test data
    data = np.array([[[1, np.nan], [3, 4], [5, 6]], [[7, np.nan], [9, 10], [11, 12]]])
    result = logsumexp_agg(data)
    assert result.shape == (2, 3)


def test_make_multivec():
    # Create a test dataframe for chr1 and chr2 that is 5 mb long
    df = pd.DataFrame(
        {
            "chrom": ["chr1"] * 10000 + ["chr2"] * 10000,
            "start": list(range(0, 5000000, 500)) * 2,
            "end": list(range(500, 5000500, 500)) * 2,
            "feature1": np.random.rand(20000),
            "feature2": np.random.rand(20000) * 100 + 3,
        }
    )
    feature_names = ["feature1", "feature2"]
    base_res = 500
    chrom_sizes = {"chr1": 5000500, "chr2": 5000500}
    outpath = f"test-{uuid4()!s}.h5"

    try:
        make_multivec(outpath, df, feature_names, base_res, chrom_sizes)
        assert isfile(outpath), "Output file was not created"

        with h5py.File(outpath, "r") as f:
            # Verify metadata
            assert f["info"].attrs["tile-size"] == 1024, "Tile-size metadata is incorrect"
            chrom_names = f["chroms"]["name"][:]
            chrom_lengths = f["chroms"]["length"][:]
            assert len(chrom_names) == 2, "Incorrect number of chromosomes"
            assert all(name.decode() in chrom_sizes for name in chrom_names), "Chromosome names mismatch"
            assert all(length == chrom_sizes[name.decode()] for name, length in zip(chrom_names, chrom_lengths)), (
                "Chromosome lengths mismatch"
            )

            # Check data structure
            assert "resolutions" in f, "Resolutions group is missing"
            res_group = f["resolutions"][str(base_res)]
            assert set(feature_names).issubset(set(res_group.attrs["row_infos"].astype(str))), (
                "Feature names mismatch in attributes"
            )

            # Validate data integrity for the highest resolution level
            for chrom in chrom_sizes.keys():
                dset = res_group["values"][str.encode(chrom)]
                # Ensure dataset shape matches expected shape based on input DataFrame and chrom_sizes
                expected_shape = (chrom_sizes[chrom] // base_res - 1, len(feature_names))
                assert dset.shape == expected_shape, f"Dataset shape for {chrom} is incorrect"
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    finally:
        remove(outpath) if isfile(outpath) else None



"""Utility functions for working with genomic bins."""

import os

import bioframe
import numpy as np
import pandas as pd

# Set the cache directory, default to ~/cache if not set
CACHE_DIR = os.getenv("CACHE_DIR", os.path.expanduser("~/cache"))


def distance_traveled_by_bin(
    df: pd.DataFrame, time_column: str, time_points: list[str], prefix: str = "PCA"
) -> pd.DataFrame:
    """Calculate the distance traveled by each bin in the PCA space by each bin across time points.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with PCA coordinates for each bin at each time point.
    time_column : str
        Column name for the time points.
    time_points : List[str]
        List of time points to calculate distance traveled.
    prefix : str
        Prefix for the component columns.

    """
    # Filter rows to only those in time_points
    df = df.dropna().copy()

    # Pivot the dataframe to have one row per bin and one column per time point
    df = df.pivot(index=["chrom", "start"], columns=time_column, values=[c for c in df.columns if c.startswith(prefix)])

    # Calculate the distance between each time point
    distance_columns = []
    for i in range(1, len(time_points)):
        # All components for the first time point
        start = df.xs(time_points[i - 1], axis=1, level=1)
        # All components for the second time point
        end = df.xs(time_points[i], axis=1, level=1)
        # Calculate the distance between the two time points
        distance = np.linalg.norm(start - end, axis=1)
        # Add the distance to the dataframe
        df[f"{time_points[i - 1]}_{time_points[i]}_distance"] = distance
        distance_columns.append(f"{time_points[i - 1]}_{time_points[i]}_distance")

    # Sum the distance across all components
    df["total_distance"] = df[distance_columns].sum(axis=1)

    # Return the chrom, start, and total distance
    df = df[["total_distance"]].reset_index()

    # Reset column name (total_distance, None) to total_distance
    df.columns = df.columns.get_level_values(0)

    return df


def make_chromarms(chromsizes, mids, binsize=None, suffixes=("p", "q")):
    """Split chromosomes into chromosome arms.

    Parameters
    ----------
    chromsizes : pandas.Series
        Series mapping chromosomes to lengths in bp.
    mids : dict-like
        Mapping of chromosomes to midpoint locations.
    binsize : int, optional
        Round midpoints to nearest bin edge for compatibility with a given
        bin grid.
    suffixes : tuple, optional
        Suffixes to name chromosome arms. Defaults to p and q.

    Returns
    -------
    4-column BED-like DataFrame (chrom, start, end, name).
    Arm names are chromosome names + suffix.
    Any chromosome not included in ``mids`` will be omitted.

    """
    chromosomes = [chrom for chrom in chromsizes.index if chrom in mids]

    p_arms = [[chrom, 0, mids[chrom], chrom + suffixes[0]] for chrom in chromosomes]
    if binsize is not None:
        for x in p_arms:
            x[2] = round(x[2] / binsize) * binsize

    q_arms = [[chrom, mids[chrom], chromsizes[chrom], chrom + suffixes[1]] for chrom in chromosomes]
    if binsize is not None:
        for x in q_arms:
            x[1] = round(x[1] / binsize) * binsize

    interleaved = [*sum(zip(p_arms, q_arms), ())]

    return pd.DataFrame(interleaved, columns=["chrom", "start", "end", "name"])


def assign_arm_labels(df: pd.DataFrame, arms: pd.DataFrame) -> pd.DataFrame:
    """Assign chromosome arm labels to bins in a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with chrom, start, and end columns.
    arms : pd.DataFrame
        DataFrame with chrom, start, end, and name columns.

    Returns
    -------
    pd.DataFrame
        DataFrame with arm column added.

    """
    g = {
        arm["name"]: bioframe.select(df, (arm.chrom, arm.start, arm.end)).assign(arm=arm["name"])
        for _, arm in arms.iterrows()
    }
    return pd.concat(g.values(), ignore_index=True)


def assign_centel_distance(group: pd.DataFrame, arms: pd.DataFrame) -> pd.DataFrame:
    """Assign centromere-proximal distance to bins in a chromosome arm.

    Parameters
    ----------
    group : pd.DataFrame
        DataFrame with chrom, start, and end columns.
    arms : pd.DataFrame
        DataFrame with chrom, start, end, and name columns.

    Returns
    -------
    pd.DataFrame
        DataFrame with centel column added.

    """
    arms = arms.set_index("name")
    this_arm = group.name
    if group.name.endswith("p"):
        arm_len = arms.loc[this_arm, "end"]
        return 1 - (group["end"] / arm_len)
    elif group.name.endswith("q"):
        arm_start = arms.loc[this_arm, "start"]
        arm_len = arms.loc[this_arm, "end"] - arm_start
        return (group["end"] - arm_start) / arm_len
    else:
        return group.assign(dummy=np.nan)["dummy"]


def centel_distance(assembly: str = "hg38", binsize: int = 50000) -> pd.DataFrame:
    """Calculate centromere-proximal distance for each bin in the genome.

    Parameters
    ----------
    assembly : str
        Genome assembly to use. Default is 'hg38'.
    binsize : int
        Bin size in bp. Default is 50,000.

    """
    # Check for cached results
    cache_path = f"{CACHE_DIR}/centel_distance_{assembly}_{binsize}.parquet"
    if os.path.exists(cache_path):
        return pd.read_parquet(cache_path)

    chrom_sizes = bioframe.fetch_chromsizes(assembly)
    # SHORT_ARMS: list = ['chr13p', 'chr14p', 'chr15p', 'chr21p', 'chr22p', 'chrYp', 'chrYq']

    # Grab centromere locations and make a chromosome arms dataframe
    centros = bioframe.fetch_centromeres("hg38").set_index("chrom")
    arms = make_chromarms(chrom_sizes, centros["mid"], binsize)

    # Calculate mapping of chromosome arm labels to their lengths
    armlens: dict = arms.assign(length=arms["end"] - arms["start"]).set_index("name")["length"].to_dict()

    # Make a bin table
    df = bioframe.binnify(chrom_sizes, binsize)

    # Calculate GC content
    # fa_records = bioframe.load_fasta('/net/levsha/share/lab/genomes/hg38/hg38.fa')
    # df = bioframe.frac_gc(df, fa_records)

    # To each bin, assign arm label, arm length, and relative/absolute centel distance
    df = assign_arm_labels(df, arms)
    df["armlen"] = df["arm"].apply(armlens.get)
    df["centel"] = df.groupby("arm", sort=False).apply(assign_centel_distance, arms=arms).reset_index(drop=True)
    df["centel_abs"] = np.round(df["centel"] * df["armlen"]).astype(int)

    # Save bin table for caching
    df.to_parquet(cache_path)
    return df



import argparse

import pytest
import yaml

from jointly_hic.hdf5db.trackcsv2yaml import csv_to_yaml


def test_csv_to_yaml(tmp_path):
    # Prepare a mock CSV input
    csv_content = """biosample1,assay1,experiment1,accession1
biosample2,assay2,experiment2,accession2
biosample3,assay3,experiment3,accession3"""
    csv_file = tmp_path / "input.csv"
    csv_file.write_text(csv_content)

    # Define the expected YAML output
    expected_metadata = [
        {"biosample": "biosample1", "assay": "assay1", "experiment": "experiment1", "accession": "accession1"},
        {"biosample": "biosample2", "assay": "assay2", "experiment": "experiment2", "accession": "accession2"},
        {"biosample": "biosample3", "assay": "assay3", "experiment": "experiment3", "accession": "accession3"},
    ]

    # Define the output YAML file
    yaml_file = tmp_path / "output.yaml"

    # Call the function
    args = argparse.Namespace(csv_file=csv_file, yaml_file=yaml_file)
    csv_to_yaml(args)

    # Read and assert the YAML output
    with open(yaml_file, "r") as file:
        content = yaml.safe_load(file)

    assert content == expected_metadata


def test_csv_to_yaml_empty_file(tmp_path):
    # Prepare an empty CSV file
    csv_file = tmp_path / "empty.csv"
    csv_file.write_text("")

    # Define the output YAML file
    yaml_file = tmp_path / "empty_output.yaml"

    # Call the function
    args = argparse.Namespace(csv_file=csv_file, yaml_file=yaml_file)
    csv_to_yaml(args)

    # Read and assert the YAML output
    with open(yaml_file, "r") as file:
        content = yaml.safe_load(file)

    assert content == []


def test_csv_to_yaml_incorrect_format(tmp_path):
    # Prepare a CSV file with incorrect format (missing one column)
    csv_content = """biosample1,assay1,experiment1
biosample2,assay2"""
    csv_file = tmp_path / "incorrect.csv"
    csv_file.write_text(csv_content)

    # Define the output YAML file
    yaml_file = tmp_path / "incorrect_output.yaml"

    # Call the function and expect it to raise an error
    with pytest.raises(ValueError):
        # Call the function
        args = argparse.Namespace(csv_file=csv_file, yaml_file=yaml_file)
        csv_to_yaml(args)



"""Functions for creating heatmaps of ENCODE ChIP-seq signal."""

import matplotlib.pyplot as plt
import pandas as pd

from jointly_hic.notebook_utils.encode_utils import EncodeFile, get_signal


def encode_chip_heatmap(
    bins: pd.DataFrame, data_cols: list, name: str, bigwigs: dict, labels: list, output_prefix: str | None = None
):
    """Create a heatmap of ENCODE ChIP-seq signal for a given search term.

    Parameters
    ----------
    bins : pd.DataFrame
        The bins to plot
    data_cols : list
        The list of columns containing the data to plot
    name : str
        The name of the dataset to plot, since bins contains seeveral datasets
    bigwigs : dict["label": "accession"]
        A dictionary of encode bigwigs with the label as the key and the accession as the value.
    labels : str
        The name of the column containing the cluster labels
    output_prefix : str
        The prefix to use for the output files

    """
    df = bins.loc[bins.name == name, :].dropna().copy()

    # Get the signal for each bigwig
    for label, accession in bigwigs.items():
        with EncodeFile(accession, "bw") as bigwig:
            df[label] = get_signal(bigwig, df)
            data_cols.append(label)

    # Clip at 1% and 99% and Z-score data columns
    for col in data_cols:
        df[col] = df[col].clip(lower=df[col].quantile(0.01), upper=df[col].quantile(0.99))
        df[col] = (df[col] - df[col].mean()) / df[col].std()

    # Convert label columns to Categorical
    for label in labels:
        df[label] = pd.Categorical(df[label])

    # Sort dataframe
    df = df.sort_values([*labels, "chrom", "start"]).reset_index(drop=True)

    # Save dataframe to parquet
    if output_prefix is not None:
        df.to_parquet(f"{output_prefix}.pq")

    # Set up plot
    fig = plt.figure(figsize=(10, 5))
    fig.subplots_adjust(wspace=0.15)
    gs = plt.GridSpec(
        nrows=1 + len(data_cols),
        ncols=1,
        height_ratios=[2] * len(labels) + [1] * len(data_cols),
        hspace=0,
    )

    # Plot labels
    for i, label in enumerate(labels):
        # Plot label values
        ax = plt.subplot(gs[i])
        ax.imshow(
            df[label].to_numpy().reshape(1, -1),
            aspect="auto",
            cmap="tab20",
            rasterized=True,
        )
        ax.set_ylabel(label, rotation="horizontal", ha="right", va="center")
        ax.set_aspect("auto")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.xaxis.set_visible(False)

    # Plot signals
    for i, signal in enumerate(data_cols):
        ax = plt.subplot(gs[i + len(labels)])
        ax.imshow(
            df[signal].to_numpy().reshape(1, -1),
            aspect="auto",
            cmap="seismic",
            rasterized=True,
        )
        y_label = ".".join(signal.split(".")[0:2])
        ax.set_ylabel(y_label, rotation="horizontal", ha="right", va="center")
        ax.set_xticks([])
        ax.set_yticks([])

    # Save figure
    if output_prefix is not None:
        fig.savefig(f"{output_prefix}_heatmap.png", dpi=600, bbox_inches="tight")

    return df



"""Utilities to create multivec files."""

from math import ceil, log

import numpy as np
from h5py import File as h5py_File
from scipy.special import logsumexp


def nansum_agg(x):
    """Sum the values in x, ignoring NaNs."""
    return np.nansum(x.T.reshape((x.shape[1], -1, 2)), axis=2).T


def nanmean_agg(x):
    """Calculate the mean of the values in x, ignoring NaNs."""
    return np.nanmean(x.T.reshape((x.shape[1], -1, 2)), axis=2).T


def logsumexp_agg(x):
    """Calculate the log sum of the exponentials of the values in x, ignoring NaNs."""
    a = x.T.reshape((x.shape[1], -1, 2))
    return logsumexp(a, axis=2).T


def make_multivec(
    outpath,
    df,
    feature_names,
    base_res,
    chromsizes,
    tilesize=1024,
    agg=nansum_agg,
    chunksize=int(1e5),
    h5opts=None,
):
    """Create a multivec file from a pandas DataFrame.

    Parameters
    ----------
    outpath : str
        Path to the output file.
    df : pandas.DataFrame
        The dataframe containing the data to be stored.
    feature_names : list of str
        The names of the columns in the dataframe to be stored.
    base_res : int
        The resolution of the data in the dataframe.
    chromsizes : dict
        A dictionary-like object mapping chromosome names to their lengths.
    tilesize : int
        The size of the tiles in the output file.
    agg : function
        The function to use to aggregate data when creating lower-resolution
        zoom levels. The function should take a 2D array of shape (n, m) and
        return a 2D array of shape (n, m // 2).
    chunksize : int
        The size of the chunks to use when writing the data to the output file.
    h5opts : dict
        A dictionary of options to pass to h5py when creating the output file.

    """
    if h5opts is None:
        h5opts = {"compression": "gzip", "compression_opts": 6, "shuffle": True}

    chromosomes = list(chromsizes.keys())
    grouped = df.groupby("chrom")
    array_dict = {chrom: grouped.get_group(chrom).loc[:, feature_names].values for chrom in chromosomes}
    chroms, lengths = zip(*chromsizes.items())
    chrom_array = np.array(chroms, dtype="S")
    feature_names = np.array(feature_names, dtype="S")

    # this will be the file that contains our multires data
    with h5py_File(outpath, "w") as f:
        # store some metadata
        f.create_group("info")
        f["info"].attrs["tile-size"] = tilesize
        f.create_group("chroms")
        f["chroms"].create_dataset(
            "name",
            shape=(len(chroms),),
            dtype=chrom_array.dtype,
            data=chrom_array,
            **h5opts,
        )
        f["chroms"].create_dataset("length", shape=(len(chroms),), data=lengths, **h5opts)

        # the data goes here
        f.create_group("resolutions")
        # the maximum zoom level corresponds to the number of aggregations
        # that need to be performed so that the entire extent of
        # the dataset fits into one tile
        total_length = sum(lengths)
        max_zoom = ceil(log(total_length / (tilesize * base_res)) / log(2))

        # start with a resolution of 1 element per pixel
        res = base_res
        grp = f["resolutions"].create_group(str(res))
        # add information about each of the rows
        if feature_names is not None:
            grp.attrs["row_infos"] = feature_names
        # hard links
        grp.create_group("chroms")
        grp["chroms"]["name"] = f["chroms"]["name"]
        grp["chroms"]["length"] = f["chroms"]["length"]
        # add the data
        grp.create_group("values")
        for chrom, length in zip(chroms, lengths):
            if chrom not in array_dict:
                continue

            dset = grp["values"].create_dataset(str(chrom), array_dict[chrom].shape, **h5opts)
            start = 0
            step = int(min(chunksize, len(dset)))
            while start < len(dset):
                # see above section
                dset[start : start + step] = array_dict[chrom][start : start + step]
                start += int(min(chunksize, len(array_dict[chrom]) - start))

        # we're going to go through and create the data for the different
        # zoom levels by summing adjacent data points
        prev_res = res
        for i in range(max_zoom):
            # each subsequent zoom level will have half as much data
            # as the previous
            res = prev_res * 2
            prev_grp, grp = grp, f["resolutions"].create_group(str(res))
            # add information about each of the rows
            if feature_names is not None:
                grp.attrs["row_infos"] = feature_names
            # hard links
            grp.create_group("chroms")
            grp["chroms"]["name"] = f["chroms"]["name"]
            grp["chroms"]["length"] = f["chroms"]["length"]
            # add the data
            grp.create_group("values")
            for chrom, length in zip(chroms, lengths):
                if chrom not in prev_grp["values"]:
                    continue

                prev_dset = prev_grp["values"][chrom]
                start = 0
                step = int(min(chunksize, len(prev_dset)))

                shape = (ceil(prev_dset.shape[0] / 2), prev_dset.shape[1])
                dset = grp["values"].create_dataset(chrom, shape, **h5opts)
                while start < len(prev_dset):
                    prev_data = prev_dset[start : start + step]

                    # this is a sort of roundabout way of calculating the
                    # shape of the aggregated array, but all its doing is
                    # just halving the first dimension of the previous shape
                    # without taking into account the other dimensions
                    if len(prev_data) % 2 != 0:
                        # we need our array to have an even number of elements
                        # so we just add the last element again
                        prev_data = np.concatenate((prev_data, [prev_data[-1]]))
                        step += 1

                    data = agg(prev_data)
                    dset[int(start / 2) : int(start / 2 + step / 2)] = data
                    start += int(min(chunksize, len(prev_dset) - start))

            prev_res = res



import os
from uuid import uuid4

import pytest
import bioframe
import numpy as np

from jointly_hic.notebook_utils.plotting import plot_components, plot_signal, create_dotplot


@pytest.fixture
def bins():
    bins = bioframe.binnify(bioframe.fetch_chromsizes("hg38"), binsize=1000000)
    bins["PCA1"] = np.random.rand(len(bins))
    bins["PCA2"] = np.random.rand(len(bins))
    bins["good_bin"] = [True] * len(bins)
    bins["name"] = np.random.choice(["sample1", "sample2"], len(bins))
    bins["cluster"] = np.random.choice([1, 2, 3], len(bins))
    return bins


def test_plot_components(bins):
    filename = f"{uuid4()!s}.png"
    plot_components(
        bins,
        component_1="PCA1",
        component_2="PCA2",
        cluster_col="cluster",
        name_col="name",
        filename=filename,
        remove_ticks=True,
    )
    os.remove(filename)


def test_plot_signal(bins):
    filename = f"{uuid4()!s}.png"
    plot_signal(
        bins,
        component_1="PCA1",
        component_2="PCA2",
        signals=["PCA1"],
        labels=["cluster"],
        filename=filename,
        remove_ticks=True,
    )
    os.remove(filename)


def test_create_dotplot(bins):
    filename = f"{uuid4()!s}.png"
    create_dotplot(bins, "cluster", ["PCA1", "PCA2"], filename=filename, threshold=0.5)
    os.remove(filename)



"""Configuration for jointly_hic."""

from datetime import datetime

from pydantic import BaseModel, Field


class JointlyConfiguration(BaseModel):
    """Configuration for Jointly.

     Run PCA or NMF on Hi-C contact matrices (in mcool format) from multiple samples using
        incremental embedding algorithm.

    Parameters
    ----------
    mcools: List[str]
        Path to mcool file(s) containing Hi-C contact matrices
    resolution: int
        Bin size to use for contact frequency matrix
    assembly: str
        Genome assembly name used for alignment (e.g. hg38, mm10)
    output: str
        Prefix for output files (Default: output_<datetime>)
    components: int
        Number of components to use for PCA (Default: 32)
    chrom_limit: int
        Limit number of chromosomes to load (Example: '23' to limit to human chr1-chrX and exclude chrY and chrM).
    method: str
        Method to use for decomposition, either PCA or NMF (Default: PCA)
    percentile_top: float
        Top percentile to filter based off of (Default: 99.5)
    percentile_bottom: float
        Bottom percentile to filter based off of (Default: 1.0)
    batch_size: int
        Batch size for mini-batches (Default: 10000)
    log_level: str
        Logging level verbosity (Default: INFO)

    """

    mcools: list[str] = Field(description="Path to mcool file(s) containing Hi-C contact matrices")
    resolution: int = Field(description="Bin size to use for contact frequency matrix")
    assembly: str = Field(description="Genome assembly name used for alignment (e.g. hg38, mm10)")
    output: str = Field(
        default_factory=lambda: f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        description="Prefix for output files",
    )
    components: int = Field(default=32, description="Number of components to use for PCA")
    chrom_limit: int = Field(default=-1, description="Limit number of chromosomes to load.")
    method: str = Field(default="PCA", description="Method to use for PCA")
    exclusion_list: str | None = Field(default=None, description="File containing regions to exclude from analysis")
    percentile_top: float = Field(default=99.5, description="Top percentile to filter based off of")
    percentile_bottom: float = Field(default=1.0, description="Bottom percentile to filter based off of")
    batch_size: int = Field(default=10000, description="Batch size for mini-batches")
    log_level: str = Field(default="INFO", description="Logging level verbosity")

    def __str__(self):
        """Pretty print configuration with an indented line per field."""
        result = "====================  Joint PCA Configuration  ====================\n"
        config_dict = self.model_dump()

        max_key_length = max(len(key) for key in config_dict.keys())
        for key, value in config_dict.items():
            if isinstance(value, list):
                value = "\n".join(f"\t\t\t\t- {v}" for v in value)
                result += f"\t{key: <{max_key_length}}  :\n{value}\n"
            else:
                result += f"\t{key: <{max_key_length}}  :\t{value}\n"
        result += "===================================================================\n"
        return result


class PostProcessingConfig(BaseModel):
    """Configuration for post-processing.

    Parameters
    ----------
    parquet_file: str
        Path to parquet file containing embedding results
    output: str
        Prefix for output files (Default: output_<datetime>)
    umap_neighbours: List[int]
        Number of neighbours to use for UMAP
    kmeans_clusters: List[int]
        Number of clusters to use for KMeans
    leiden_resolutions: List[float]
        Resolution to use for Leiden clustering
    method: str
        Method used for decomposition
    log_level: str
        Logging level verbosity (Default: INFO)

    """

    parquet_file: str = Field(description="Path to parquet file containing embedding results")
    output: str = Field(
        default_factory=lambda: f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        description="Prefix for output files",
    )
    umap_neighbours: list[int] = Field(default=[30, 100, 500], description="Number of neighbours to use for UMAP")
    kmeans_clusters: list[int] = Field(
        default=[5, 6, 7, 8, 9, 10, 15, 20], description="Number of clusters to use for KMeans"
    )
    leiden_resolutions: list[float] = Field(
        default=[0.1, 0.2, 0.3, 0.5, 0.8, 1.0], description="Resolution to use for Leiden clustering"
    )
    method: str = Field(default="PCA", description="Method to use for decomposition")
    log_level: str = Field(default="INFO", description="Logging level verbosity")

    def __str__(self):
        """Pretty print configuration with an indented line per field."""
        result = "====================  Post Processor Configuration  ====================\n"
        config_dict = self.model_dump()

        max_key_length = max(len(key) for key in config_dict.keys())
        for key, value in config_dict.items():
            result += f"\t{key: <{max_key_length}}  :\t{value}\n"
        result += "========================================================================\n"
        return result


class TrajectoryAnalysisConfig(BaseModel):
    """Configuration for trajectory analysis.

    Parameters
    ----------
    parquet_file: str
        Path to parquet file containing embedding results
    output: str
        Prefix for output files (Default: output_<datetime>)
    umap_neighbours: List[int]
        Number of neighbours to use for UMAP
    kmeans_clusters: List[int]
        Number of clusters to use for KMeans
    log_level: str
        Logging level verbosity (Default: INFO)

    """

    parquet_file: str = Field(description="Path to parquet file containing embedding results")
    output: str = Field(
        default_factory=lambda: f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        description="Prefix for output files",
    )
    kmeans_clusters: list[int] = Field(
        default=[5, 6, 7, 8, 9, 10, 15, 20], description="Number of clusters to use for KMeans"
    )
    leiden_neighbors: int = Field(default=100, description="Number of neighbors to use for Leiden clustering")
    umap_neighbours: list[int] = Field(default=[30, 100, 500], description="Number of neighbours to use for UMAP")
    method: str = Field(default="PCA", description="Method to use for decomposition")
    log_level: str = Field(default="INFO", description="Logging level verbosity")

    def __str__(self):
        """Pretty print configuration with an indented line per field."""
        result = "====================  Trajectory Analysis Configuration  ====================\n"
        config_dict = self.model_dump()

        max_key_length = max(len(key) for key in config_dict.keys())
        for key, value in config_dict.items():
            result += f"\t{key: <{max_key_length}}  :\t{value}\n"
        result += "=============================================================================\n"
        return result



"""Utilties to convert a dataframe to a bed9 file with clusters."""

import bioframe
import click
import numpy as np
import pandas as pd
import seaborn as sns
from bioframe import ops
from bioframe.core.specs import _get_default_colnames


def mark_runs(
    df: pd.DataFrame,
    col: str,
    *,
    allow_overlaps: bool = False,
    reset_counter: bool = True,
    run_col: str = "run",
    cols: tuple[str, str, str] | None = None,
) -> pd.DataFrame:
    """Mark runs of immediately consecutive intervals.

    Parameters
    ----------
    df : DataFrame
        A bioframe dataframe.
    col : str
        The column to mark runs of values for.
    allow_overlaps : bool, optional [default: False]
        If True, allow intervals in ``df`` to overlap. This may cause
        unexpected results.
    reset_counter : bool, optional [default: True]
        If True, reset the run counter for each chromosome.
    run_col : str, optional [default: 'run']
        The name of the column to store the run numbers in.
    cols : tuple of str, optional
        The names of the columns in the input dataframe to use. If not
        provided, the default column names are used.

    Returns
    -------
    pandas.DataFrame
        A reordered copy the input dataframe with an additional column 'run'
        marking runs of values in the input column.

    See Also
    --------
    merge_runs

    """
    ck, sk, ek = _get_default_colnames() if cols is None else cols

    if not allow_overlaps and len(ops.overlap(df, df)) > len(df):
        raise ValueError("Not a proper bedGraph: found overlapping intervals.")

    result = []
    n_runs = 0

    for _, group in df.groupby(ck, sort=False):
        group = group.sort_values([sk, ek])
        starts = group[sk].to_numpy()
        ends = group[ek].to_numpy()

        # Extend ends by running max
        ends = np.maximum.accumulate(ends)

        # Find borders of interval clusters and assign cluster ids
        is_cluster_border = np.r_[True, starts[1:] > ends[:-1], False]

        # Find borders of consecutive equal values
        values = group[col].to_numpy()
        if values.dtype.kind == "f":
            is_value_border = np.r_[True, ~np.isclose(values[1:], values[:-1], equal_nan=True), False]
        else:
            is_value_border = np.r_[True, values[1:] != values[:-1], False]

        # Find index extents of runs
        is_border = is_cluster_border | is_value_border
        sum_borders = np.cumsum(is_border)
        run_ids = sum_borders[:-1] - 1

        # Assign run numbers to intervals
        if reset_counter:
            n_runs = 0
        group[run_col] = n_runs + run_ids
        n_runs += sum_borders[-1]

        result.append(group)

    return pd.concat(result)


def merge_runs(
    df: pd.DataFrame,
    col: str,
    *,
    allow_overlaps: bool = False,
    agg: dict | None = None,
    cols: tuple[str, str, str] | None = None,
) -> pd.DataFrame:
    """Merge runs of immediately consecutive intervals sharing the same value.

    Parameters
    ----------
    df : DataFrame
        A bioframe dataframe.
    col : str
        The column to compress runs of values for.
    allow_overlaps : bool, optional [default: False]
        If True, allow intervals in ``df`` to overlap. This may cause
        unexpected results.
    agg : dict, optional [default: None]
        A dictionary of additional column names and aggregation functions to
        apply to each run. Takes the format:
            {'agg_name': ('column_name', 'agg_func')}
    cols : tuple of str, optional
        The names of the columns in the input dataframe to use. If not
        provided, the default column names are used.

    Returns
    -------
    pandas.DataFrame
        Dataframe with consecutive intervals in the same run merged.

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     'chrom': ['chr1', 'chr1', 'chr1', 'chr1', 'chr1', 'chr1'],
    ...     'start': [0, 100, 200, 300, 400, 500],
    ...     'end': [100, 200, 300, 400, 500, 600],
    ...     'value': [1, 1, 1, 2, 2, 2],
    ... })

    >>> merge_runs(df, 'value')
        chrom  start  end  value
    0   chr1      0  300      1
    1   chr1    300  600      2

    >>> merge_runs(df, 'value', agg={'sum': ('value', 'sum')})
        chrom  start  end  value  sum
    0   chr1      0  300      1    3
    1   chr1    300  600      2    6

    See Also
    --------
    mark_runs

    """
    ck, sk, ek = _get_default_colnames() if cols is None else cols

    if agg is None:
        agg = {}

    df_runs = mark_runs(
        df,
        col,
        allow_overlaps=allow_overlaps,
        reset_counter=False,
        run_col="_run",
    )
    df_merged = df_runs.groupby("_run").agg(
        **{ck: (ck, "first"), sk: (sk, "min"), ek: (ek, "max"), col: (col, "first"), **agg}
    )
    return df_merged.reset_index(drop=True)


@click.command()
@click.option(
    "--input_file",
    "-i",
    required=True,
    help="Input file (parquet, pq or csv(.gz))",
)
@click.option(
    "--output_prefix",
    "-o",
    required=True,
    help="Output file prefix",
)
@click.option(
    "--cluster_col",
    "-c",
    default="leiden",
    help="Column name for clusters",
)
@click.option(
    "--palette",
    "-p",
    default="tab20",
    help="Color palette",
)
@click.option(
    "--name_column",
    "-n",
    default="filename",
    help="Column name for filenames",
)
def create_bed9_clusters(
    input_file: str,
    output_prefix: str,
    cluster_col: str = "leiden",
    palette: str = "tab20",
    name_column: str = "filename",
):
    """Create bed9 files for each cluster in the input file."""
    # Read input file (can be a parquet, pq or a csv(.gz) file)
    if input_file.endswith(".pq") or input_file.endswith(".parquet"):
        df = pd.read_parquet(input_file)
    elif input_file.endswith(".csv") or input_file.endswith(".csv.gz"):
        df = pd.read_csv(input_file)
    else:
        raise ValueError(f"Input file format not supported: {input_file}")

    # Convert start and end to integers, drop NAs, convert cluster_col to Categorical
    df = df.dropna(subset=cluster_col)
    df["start"] = df["start"].astype(int)
    df["end"] = df["end"].astype(int)
    df[cluster_col] = df[cluster_col].astype(int).astype("category")

    # Create palette
    palette = sns.color_palette(palette, len(df[cluster_col].unique()))
    palette_mapping = {k: v for k, v in zip(df[cluster_col].unique(), palette)}
    colors = {k: bioframe.to_ucsc_colorstring(v) for k, v in palette_mapping.items()}

    # Iterate through all name_columns in df and create bed9 files
    for i, dataset_df in df.groupby(name_column):
        merged_df = merge_runs(dataset_df, cluster_col)
        result = pd.DataFrame(
            {
                "chrom": merged_df["chrom"],
                "start": merged_df["start"],
                "end": merged_df["end"],
                "name": merged_df[cluster_col],
                "score": 0,
                "strand": ".",
                "thickStart": merged_df["start"],
                "thickEnd": merged_df["end"],
                "itemRgb": merged_df[cluster_col].map(colors),
            }
        )
        result.to_csv(f"{output_prefix}_{i}.bed.gz", sep="\t", header=False, index=False)


# Create CLI entrypoint for create_bed9_clusters
if __name__ == "__main__":
    create_bed9_clusters()



import numpy as np
import pytest
import scipy
from sklearn.decomposition import TruncatedSVD
from sklearn.utils import gen_batches

from jointly_hic.methods.sparse_incremental_svd import SparseIncrementalSVD

N_COMPONENTS = 20
BATCH_SIZE = 50
MATRIX_SIZE = 200


@pytest.fixture(scope="module")
def toeplitz():
    mat = scipy.linalg.toeplitz(np.arange(MATRIX_SIZE)).astype(np.float64)
    return mat


@pytest.fixture(scope="module")
def sparse_toeplitz(toeplitz):
    return scipy.sparse.csr_matrix(toeplitz)


@pytest.fixture(scope="module")
def n_samples(sparse_toeplitz):
    return sparse_toeplitz.shape[0]


def truncated_svd(matrix, n_components=N_COMPONENTS):
    svd = TruncatedSVD(n_components=n_components)
    svd = svd.fit(matrix)
    return svd


def sparse_incremental_svd(matrix, n_components=N_COMPONENTS):
    svd = SparseIncrementalSVD(n_components=n_components, batch_size=BATCH_SIZE)
    svd = svd.partial_fit(matrix)
    return svd


def test_truncated_svd(toeplitz):
    svd = truncated_svd(toeplitz)
    assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_sparse_truncated_svd(sparse_toeplitz):
    svd = truncated_svd(sparse_toeplitz)
    assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_incremental_svd_partial_fit(toeplitz):
    svd = SparseIncrementalSVD(n_components=N_COMPONENTS, batch_size=BATCH_SIZE)
    svd.partial_fit(toeplitz)
    assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_sparse_incremental_svd_partial_fit(sparse_toeplitz):
    svd = sparse_incremental_svd(sparse_toeplitz)
    assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_sparse_incremental_svd_fit(sparse_toeplitz):
    svd = SparseIncrementalSVD(n_components=N_COMPONENTS, batch_size=BATCH_SIZE)
    svd.fit(sparse_toeplitz)
    assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_incremental_svd_minibatch_fit(sparse_toeplitz, n_samples):
    svd = SparseIncrementalSVD(n_components=N_COMPONENTS, batch_size=BATCH_SIZE)
    for i, batch in enumerate(gen_batches(n_samples, BATCH_SIZE, min_batch_size=N_COMPONENTS)):
        svd.partial_fit(sparse_toeplitz[batch])
        if i == 0:
            assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)
        else:
            assert svd.components_.shape == (N_COMPONENTS, MATRIX_SIZE)


def test_same_results(sparse_toeplitz):
    """Test that SparseIncrementalSVD and TruncatedSVD give the same results (R^2>0.999)."""
    svd = truncated_svd(sparse_toeplitz)
    svd2 = sparse_incremental_svd(sparse_toeplitz)

    # Assert components_ are correlated > 0.999
    for i in range(N_COMPONENTS):
        corr = np.corrcoef(svd.components_[i], svd2.components_[i])[0, 1]
        assert np.abs(corr) > 0.999

    assert np.allclose(np.abs(svd.singular_values_), np.abs(svd2.singular_values_), rtol=1e-3)


def test_transform(sparse_toeplitz):
    svd = sparse_incremental_svd(sparse_toeplitz, N_COMPONENTS)
    transformed = svd.transform(sparse_toeplitz)
    assert transformed.shape == (MATRIX_SIZE, N_COMPONENTS)



"""Parser for jointly_hic."""

import argparse
from datetime import datetime

from jointly_hic.core.config import JointlyConfiguration, PostProcessingConfig, TrajectoryAnalysisConfig
from jointly_hic.hdf5db.experiments2yaml import configure_embedding_metadata_parser
from jointly_hic.hdf5db.hdf5db import configure_hdf5db_parser
from jointly_hic.hdf5db.trackcsv2yaml import configure_trackcsv2yaml_parser


class JointlyCommandLineInterface:
    """Main CLI for jointly_hic."""

    def __init__(self):
        """Initialize parser."""
        parser = argparse.ArgumentParser(
            prog="jointly",
            description="Jointly decompose Hi-C contact matrices using PCA or NMF and analyze the embeddings",
            epilog="For more information, visit https://github.com/abdenlab/jointly-hic",
        )
        parser.add_argument(
            "--log",
            help="Log file",
            required=False,
            type=str,
            default=f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}_log.txt",
        )
        parser.add_argument(
            "--log-level",
            help="Logging level",
            required=False,
            type=str,
            choices=["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"],
            default="INFO",
        )
        subparsers = parser.add_subparsers(help="List of commands.", required=True)

        # Embed parser
        embed_parser = subparsers.add_parser("embed", help="Embed Hi-C contact matrices using PCA or NMF.")
        self.configure_embed_parser(embed_parser)

        # Post-process parser
        post_process_parser = subparsers.add_parser("post-process", help="Post process embeddings from jointly-hic.")
        self.configure_post_process_parser(post_process_parser)

        # Trajectory parser
        trajectory_parser = subparsers.add_parser(
            "trajectory", help="Trajectory analysis of embeddings from jointly-hic."
        )
        self.configure_trajectory_parser(trajectory_parser)

        # HDF5 database parser
        hdf5db_parser = subparsers.add_parser(
            "hdf5db", help="Create HDF5 database from Jointly-hic output and signal files."
        )
        configure_hdf5db_parser(hdf5db_parser)

        # embedding-metadata parser
        embedding_metadata_parser = subparsers.add_parser(
            "embedding2yaml", help="Create metadata YAML from embeddings."
        )
        configure_embedding_metadata_parser(embedding_metadata_parser)

        # trackcsv2yaml parser
        trackcsv2yaml_parser = subparsers.add_parser(
            "tracks2yaml", help="Convert a 4-column CSV metadata file to a YAML file."
        )
        configure_trackcsv2yaml_parser(trackcsv2yaml_parser)

        self.parser = parser

    def configure_embed_parser(self, embed_parser):
        """Configure embed parser."""
        embed_parser.add_argument(
            "--mcools",
            help="Path to mcool file(s) containing Hi-C contact matrices",
            required=True,
            nargs="+",
            type=str,
        )
        embed_parser.add_argument(
            "--resolution",
            help="Bin size to use for contact frequency matrix",
            required=True,
            type=int,
        )
        embed_parser.add_argument(
            "--assembly",
            help="Genome assembly name used for alignment (e.g. hg38, mm10)",
            required=True,
            type=str,
            choices=["hg38", "mm10", "hg19", "mm9"],
        )
        embed_parser.add_argument(
            "--output",
            help="Prefix for output files (Default: output_<datetime>)",
            required=False,
            type=str,
            default=f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        )
        embed_parser.add_argument(
            "--components",
            help="Number of components to use for PCA (Default: 32)",
            required=False,
            type=int,
            default=32,
        )
        embed_parser.add_argument(
            "--chrom-limit",
            help="Limit number of chromosomes to load "
            "(Example: '23' to limit to human chr1-chrX and exclude chrY and chrM).",
            required=False,
            type=int,
            default=-1,
        )
        embed_parser.add_argument(
            "--method",
            help="Method to use for decomposition, either PCA, NMF or SVD (Default: PCA)",
            required=False,
            type=str,
            choices=["PCA", "NMF", "SVD"],
            default="PCA",
        )
        embed_parser.add_argument(
            "--exclusion-list",
            help="File containing regions to exclude from analysis",
            required=False,
            type=str,
            default=None,
        )
        embed_parser.add_argument(
            "--percentile-top",
            help="Top percentile for filtering (Default: 99.5)",
            required=False,
            type=float,
            default=99.5,
        )
        embed_parser.add_argument(
            "--percentile-bottom",
            help="Bottom percentile for filtering (Default: 1)",
            required=False,
            type=float,
            default=1,
        )
        embed_parser.add_argument(
            "--batch-size",
            help="Batch size for mini-batches (Default: 10000)",
            required=False,
            type=int,
            default=10000,
        )
        embed_parser.set_defaults(func=self.run_embedding)

    def configure_post_process_parser(self, post_process_parser):
        """Configure post-process parser."""
        post_process_parser.add_argument(
            "--parquet-file",
            help="Path to parquet file containing embeddings",
            required=True,
            type=str,
        )
        post_process_parser.add_argument(
            "--output",
            help="Prefix for output files (Default: output_<datetime>)",
            required=False,
            type=str,
            default=f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        )
        post_process_parser.add_argument(
            "--umap-neighbours",
            help="Number of neighbours for UMAP (Default: [30, 100, 500])",
            required=False,
            type=int,
            nargs="+",
            default=[30, 100, 500],
        )
        post_process_parser.add_argument(
            "--kmeans-clusters",
            help="Number of clusters for k-means (Default: [5, 10, 15, 20])",
            required=False,
            type=int,
            nargs="+",
            default=[5, 10, 15, 20],
        )
        post_process_parser.set_defaults(func=self.run_post_process)

    def configure_trajectory_parser(self, trajectory_parser):
        """Configure trajectory parser."""
        trajectory_parser.add_argument(
            "--parquet-file",
            help="Path to parquet file containing embeddings",
            required=True,
            type=str,
        )
        trajectory_parser.add_argument(
            "--output",
            help="Prefix for output files (Default: output_<datetime>)",
            required=False,
            type=str,
            default=f"output_{datetime.now().strftime('%Y_%m_%d_%H_%M')}",
        )
        trajectory_parser.add_argument(
            "--umap-neighbours",
            help="Number of neighbours for UMAP (Default: [30, 100, 500])",
            required=False,
            type=int,
            nargs="+",
            default=[30, 100, 500],
        )
        trajectory_parser.add_argument(
            "--kmeans-clusters",
            help="Number of clusters for k-means (Default: [5, 10, 15, 20])",
            required=False,
            type=int,
            nargs="+",
            default=[5, 10, 15, 20],
        )
        trajectory_parser.set_defaults(func=self.run_trajectory)

    def run_embedding(self, args):
        """Run embedding."""
        from jointly_hic.core.decomposer import JointlyDecomposer
        from jointly_hic.core.post_processor import PostProcessor
        from jointly_hic.core.trajectory_analyzer import TrajectoryAnalyzer

        # Run embedding
        configuration = JointlyConfiguration(**args.__dict__)
        decomposer = JointlyDecomposer(configuration)
        embeddings, post_config, trajectory_config = decomposer.run()

        # Run post-processing
        post_processor = PostProcessor(post_config, embeddings)
        post_processor.run()

        # Run trajectory analysis
        trajectory_analyzer = TrajectoryAnalyzer(trajectory_config, embeddings)
        trajectory_analyzer.run()

    def run_post_process(self, args):
        """Run post-processing."""
        from jointly_hic.core.post_processor import PostProcessor

        config = PostProcessingConfig(**args.__dict__)
        post_processor = PostProcessor(config)
        post_processor.run()

    def run_trajectory(self, args):
        """Run trajectory analysis."""
        from jointly_hic.core.trajectory_analyzer import TrajectoryAnalyzer

        config = TrajectoryAnalysisConfig(**args.__dict__)
        trajectory_analyzer = TrajectoryAnalyzer(config)
        trajectory_analyzer.run()



import os

import pandas as pd
import pytest
from click.testing import CliRunner

from jointly_hic.notebook_utils.bed9_conversion import mark_runs, merge_runs, create_bed9_clusters


@pytest.fixture
def sample_dataframe():
    return pd.DataFrame(
        {
            "chrom": ["chr1", "chr1", "chr2", "chr2", "chr1"],
            "start": [100, 200, 100, 200, 300],
            "end": [150, 250, 150, 250, 350],
            "value": [1, 2, 3, 4, 5],
        }
    )


def test_mark_runs_no_overlap(sample_dataframe):
    result_df = mark_runs(sample_dataframe, "value")
    assert "run" in result_df.columns


def test_merge_runs(sample_dataframe):
    merged_df = merge_runs(sample_dataframe, "value")
    assert len(merged_df) == len(sample_dataframe)


@pytest.fixture
def input_data_csv(tmpdir):
    data = """chrom,start,end,value,filename
chr1,100,150,1,file1
chr1,200,250,2,file1
chr2,100,150,3,file2
chr2,200,250,4,file2
chr1,300,350,5,file1"""
    file_path = tmpdir.join("input.csv")
    file_path.write(data)
    return str(file_path)


@pytest.fixture
def runner():
    return CliRunner()


def test_create_bed9_clusters_csv_input(runner, input_data_csv, tmpdir):
    output_prefix = str(tmpdir.join("output"))
    result = runner.invoke(
        create_bed9_clusters, ["-i", input_data_csv, "-o", output_prefix, "-c", "value", "-n", "filename"]
    )

    assert result.exit_code == 0
    assert os.path.exists(f"{output_prefix}_file1.bed.gz")
    assert os.path.exists(f"{output_prefix}_file2.bed.gz")


def test_unsupported_file_format(runner, tmpdir):
    unsupported_file = tmpdir.join("input.xyz")
    unsupported_file.write("dummy data")
    output_prefix = str(tmpdir.join("output"))

    result = runner.invoke(create_bed9_clusters, ["-i", str(unsupported_file), "-o", output_prefix])

    assert result.exit_code != 0
    assert isinstance(result.exception, ValueError)



import numpy as np
import pandas as pd
import pytest

from jointly_hic.core.config import PostProcessingConfig
from jointly_hic.core.post_processor import PostProcessor


@pytest.fixture
def embeddings():
    """Fixture for embeddings."""
    return pd.DataFrame(
        {
            "chrom": ["chr1"] * 1000,
            "start": np.arange(1, 1001),
            "end": np.arange(2, 1002),
            "weight": np.ones(1000),
            "good_bin": [True] * 1000,
            # Randomly create PC 1-4 as floating point columns
            "PCA1": np.random.rand(1000),
            "PCA2": np.random.rand(1000),
            "PCA3": np.random.rand(1000),
            "PCA4": np.random.rand(1000),
            "filename": ["test1.mcool"] * 1000,
        }
    )


@pytest.fixture
def configuration():
    return PostProcessingConfig(
        parquet_file="test.parquet",
        output="test",
        umap_neighbours=[30, 100],
        kmeans_clusters=[2, 3],
        trajectory_kmeans=[2],
        trajectory_umap_neighbours=[30],
        leiden_resolution=[0.5],
        log_level="INFO",
        method="PCA",
    )


@pytest.fixture
def post_processor(mocker, embeddings, configuration):
    """Fixture for PostProcessor."""
    mocker.patch("jointly_hic.core.post_processor.logger")
    # Mock all plotting and IO functions
    mocker.patch("jointly_hic.core.post_processor.PostProcessor.plot_scores")
    mocker.patch("matplotlib.pyplot.savefig")
    mocker.patch("pandas.DataFrame.to_csv")
    mocker.patch("pandas.DataFrame.to_parquet")
    return PostProcessor(configuration, embeddings)


def test_post_processor_init(mocker, embeddings, configuration):
    """Test PostProcessor.init."""
    mocker.patch("jointly_hic.core.post_processor.logger")
    post_processor = PostProcessor(configuration, embeddings)
    assert post_processor.configuration == configuration
    assert post_processor.embeddings.equals(embeddings)
    assert post_processor.cluster_columns == []


def test_post_processor_init_no_embeddings(mocker, configuration):
    """Test PostProcessor.init without embeddings."""
    mocker.patch("jointly_hic.core.post_processor.logger")
    mock_read_parquet = mocker.patch("pandas.read_parquet")
    post_processor = PostProcessor(configuration)
    assert post_processor.configuration == configuration
    assert post_processor.embeddings == mock_read_parquet.return_value
    assert post_processor.cluster_columns == []


def test_post_processor_components(mocker, post_processor, embeddings):
    """Test PostProcessor.components."""
    assert np.all(
        post_processor.components == embeddings.loc[embeddings.good_bin, ["PCA1", "PCA2", "PCA3", "PCA4"]].values
    )


def test_post_processor_add_column(mocker, post_processor, embeddings):
    """Test PostProcessor.add_column."""
    test_data = np.random.rand(embeddings.shape[0])
    post_processor.add_column("test", test_data)
    assert np.all(post_processor.embeddings["test"] == test_data)
    assert np.all(post_processor.embeddings.loc[~post_processor.embeddings.good_bin, "test"].isna())


def test_post_processor_run(mocker, post_processor):
    """Test PostProcessor.run."""
    post_processor.run_kmeans = mocker.MagicMock()
    post_processor.run_leiden = mocker.MagicMock()
    post_processor.run_umap = mocker.MagicMock()
    post_processor.plot_scores = mocker.MagicMock()

    post_processor.run()
    components = post_processor.components

    # Assert run_kmeans called with scores and n_clusters=2 and 3
    assert post_processor.run_kmeans.call_count == 2
    assert np.all(post_processor.run_kmeans.call_args_list[0][0][0] == components)
    assert post_processor.run_kmeans.call_args_list[0][1]["n_clusters"] == 2
    assert np.all(post_processor.run_kmeans.call_args_list[1][0][0] == components)
    assert post_processor.run_kmeans.call_args_list[1][1]["n_clusters"] == 3

    # Assert run_umap called with scores and n_neighbors=30 and 100
    assert post_processor.run_umap.call_count == 2
    assert np.all(post_processor.run_umap.call_args_list[0][0][0] == components)
    assert post_processor.run_umap.call_args_list[0][1]["n_neighbors"] == 30
    assert np.all(post_processor.run_umap.call_args_list[1][0][0] == components)
    assert post_processor.run_umap.call_args_list[1][1]["n_neighbors"] == 100

    # Assert plot_scores called with scores
    post_processor.plot_scores.assert_called_once()


def test_post_processor_plot_scores(mocker, configuration, embeddings):
    """Test PostProcessor.plot_scores."""
    savefig_patch = mocker.patch("matplotlib.pyplot.savefig")
    mocker.patch("datashader.mpl_ext.dsshow")
    ceil_patch = mocker.patch("numpy.ceil", return_value=2)
    dsshow_patch = mocker.patch("datashader.mpl_ext.dsshow")
    hist_patch = mocker.patch("matplotlib.pyplot.hist")
    subplots_patch = mocker.patch("matplotlib.pyplot.subplots", return_value=(mocker.MagicMock(), mocker.MagicMock()))
    post_processor = PostProcessor(configuration, embeddings)
    post_processor.plot_scores()

    # Assert savefig called with output + "_filename_scores.png" and "_clustered_scores.png"
    savefig_patch.call_count == 2
    savefig_patch.assert_any_call("test_scores.png")
    savefig_patch.assert_any_call("test_scores_filenames.png")
    ceil_patch.call_count == 2
    dsshow_patch.call_count == 2
    hist_patch.call_count == 2
    subplots_patch.call_count == 3


def test_post_processor_run_kmeans(mocker, configuration, embeddings):
    """Test PostProcessor.run_kmeans."""
    post_processor = PostProcessor(configuration, embeddings)
    post_processor.add_column = mocker.MagicMock()
    post_processor.run_kmeans(post_processor.components, n_clusters=2)

    # Assert add_column called with "kmeans_2"
    assert post_processor.add_column.call_count == 1
    assert post_processor.add_column.call_args_list[0][0][0] == "kmeans_2"
    assert post_processor.cluster_columns == ["kmeans_2"]


def test_post_processor_run_umap(mocker, configuration, embeddings):
    """Test PostProcessor.run_umap."""
    mock_umap = mocker.patch("umap.UMAP")
    savefig_patch = mocker.patch("matplotlib.pyplot.savefig")
    subplots_patch = mocker.patch("matplotlib.pyplot.subplots", return_value=(mocker.MagicMock(), mocker.MagicMock()))
    ceil_patch = mocker.patch("numpy.ceil", return_value=2)
    mocker.patch("pandas.DataFrame.loc")
    post_processor = PostProcessor(configuration, embeddings)
    post_processor.add_column = mocker.MagicMock()
    post_processor.cluster_columns = []
    # Patch dsshow and ds.Point and SclarDSArtist
    mocker.patch("datashader.mpl_ext.dsshow")
    mocker.patch("datashader.mpl_ext.ScalarDSArtist")
    mocker.patch("datashader.Point")

    post_processor.run_umap(post_processor.components, n_neighbors=30)

    mock_umap.assert_called_once_with(n_neighbors=30)
    mock_umap.return_value.fit_transform.assert_called_once_with(post_processor.components)
    assert post_processor.add_column.call_count == 2
    assert post_processor.add_column.call_args_list[0][0][0] == "umap1_n30"
    assert post_processor.add_column.call_args_list[1][0][0] == "umap2_n30"
    assert savefig_patch.call_count == 2
    assert ceil_patch.call_count == 2
    savefig_patch.assert_any_call("test_umap-n30_filenames.png")
    subplots_patch.call_count == 2


def test_run_leiden(mocker, configuration, embeddings):
    """Test PostProcessor.run_leiden."""
    post_processor = PostProcessor(configuration, embeddings)
    post_processor.add_column = mocker.MagicMock()
    post_processor.run_leiden(post_processor.components, n_neighbors=20)

    # Assert add_column called with "leiden"
    assert post_processor.add_column.call_count == 1
    assert post_processor.add_column.call_args_list[0][0][0] == "leiden_1_0_n20"
    assert post_processor.cluster_columns == ["leiden_1_0_n20"]



import numpy as np
import pytest

from jointly_hic.methods.trans_cooler_utils import normalized_affinity_matrix_from_trans


def test_normalized_affinity_matrix_from_trans(mocker):
    # Check that a non-symmetric matrix raises an error
    contact_matrix = np.array([[1, 2, 3], [4, 5, 6]])
    partition = np.array([0, 2, 3])
    percentile_top = 95
    percentile_bottom = 5
    with pytest.raises(ValueError):
        normalized_affinity_matrix_from_trans(contact_matrix, partition, percentile_top, percentile_bottom)

    # Check that a non-monotonic partition raises an error
    partition = np.array([0, 2, 1])
    with pytest.raises(ValueError):
        normalized_affinity_matrix_from_trans(contact_matrix, partition, percentile_top, percentile_bottom)

    # Check that a partition that doesn't start at 0 raises an error
    partition = np.array([1, 2])
    with pytest.raises(ValueError):
        normalized_affinity_matrix_from_trans(contact_matrix, partition, percentile_top, percentile_bottom)

    # Create an 8x8 test matrix
    contact_matrix = np.array(
        [
            [1, 2, 3, 4, 5, 6, 7, 8],
            [2, 1, 2, 3, 4, 5, 6, 7],
            [3, 2, 1, 2, 3, 4, 5, 6],
            [4, 3, 2, 1, 2, 3, 4, 5],
            [5, 4, 3, 2, 1, 2, 3, 4],
            [6, 5, 4, 3, 2, 1, 2, 3],
            [7, 6, 5, 4, 3, 2, 1, 2],
            [8, 7, 6, 5, 4, 3, 2, 1],
        ],
        dtype=np.float64,
    )
    partition = np.array([0, 4, 8])
    percentile_top = 95
    percentile_bottom = 5
    matrix, bad_bins = normalized_affinity_matrix_from_trans(
        contact_matrix, partition, percentile_top, percentile_bottom
    )

    # Check that the matrix is symmetric
    assert np.allclose(matrix, matrix.T)
    # Check that the matrix is finite
    assert np.all(np.isfinite(matrix))
    # Check that the matrix is stochastic
    assert np.allclose(np.sum(matrix, axis=0), np.ones(8), atol=0.01)
    assert np.allclose(np.sum(matrix, axis=1), np.ones(8), atol=0.01)
    # Check that the matrix is balanced
    assert np.allclose(np.sum(matrix, axis=0), np.sum(matrix, axis=1), atol=0.01)
    # Check that bad_bins is empty
    assert bad_bins.size == 8



import pytest
import pandas as pd
import yaml

from jointly_hic.hdf5db.experiments2yaml import verify_metadata_consistency, write_metadata_to_yaml, main


def test_verify_metadata_consistency():
    # Test with consistent metadata
    data = {
        "accession": ["acc1", "acc1", "acc2", "acc2"],
        "metadata1": ["value1", "value1", "value2", "value2"],
        "metadata2": ["value3", "value3", "value4", "value4"],
    }
    df = pd.DataFrame(data)
    result = verify_metadata_consistency(df, "accession", ["metadata1", "metadata2"])
    expected = [
        {"accession": "acc1", "metadata1": "value1", "metadata2": "value3"},
        {"accession": "acc2", "metadata1": "value2", "metadata2": "value4"},
    ]
    assert result == expected

    # Test with inconsistent metadata
    data = {
        "accession": ["acc1", "acc1", "acc2", "acc2"],
        "metadata1": ["value1", "value1", "value2", "value5"],
        "metadata2": ["value3", "value3", "value4", "value4"],
    }
    df = pd.DataFrame(data)
    with pytest.raises(ValueError, match="Inconsistent values found in column 'metadata1' for accession 'acc2'"):
        verify_metadata_consistency(df, "accession", ["metadata1", "metadata2"])


def test_write_metadata_to_yaml(tmp_path):
    metadata_list = [
        {"accession": "acc1", "metadata1": "value1", "metadata2": "value3"},
        {"accession": "acc2", "metadata1": "value2", "metadata2": "value4"},
    ]
    yaml_file = tmp_path / "output.yaml"
    write_metadata_to_yaml(metadata_list, yaml_file)

    with open(yaml_file, "r") as file:
        content = yaml.safe_load(file)

    assert content == metadata_list


# Mock data for end-to-end test
def test_main_end_to_end(tmp_path, monkeypatch):
    data = {
        "accession": ["acc1", "acc1", "acc2", "acc2"],
        "metadata1": ["value1", "value1", "value2", "value2"],
        "metadata2": ["value3", "value3", "value4", "value4"],
    }
    df = pd.DataFrame(data)
    parquet_file = tmp_path / "input.parquet"
    df.to_parquet(parquet_file)

    yaml_file = tmp_path / "output.yaml"

    def mock_read_parquet(file):
        return df

    monkeypatch.setattr(pd, "read_parquet", mock_read_parquet)
    main(parquet_file, "accession", ["metadata1", "metadata2"], yaml_file)

    with open(yaml_file, "r") as file:
        content = yaml.safe_load(file)

    expected = [
        {"accession": "acc1", "metadata1": "value1", "metadata2": "value3"},
        {"accession": "acc2", "metadata1": "value2", "metadata2": "value4"},
    ]

    assert content == expected



"""Utils for operating on trans- Hi-C data in cooler format."""

import logging
from itertools import pairwise

import numpy as np
from cooltools.api.eigdecomp import _fake_cis, _filter_heatmap
from cooltools.lib import numutils

logger = logging.getLogger("joint_pca")


def normalized_affinity_matrix_from_trans(
    contact_matrix, partition, percentile_top, percentile_bottom
) -> tuple[np.ndarray, np.ndarray]:
    """Normalize and balance Hi-C affinity matrix.

    Produce an affinity matrix based on trans data by filling in cis regions with randomly sampled trans pixels from
    the same row or column. The resulting matrix is rebalanced and uniformly scaled such that all rows and columns sum
    to 1 (a.k.a. a stochastic matrix).

    Parameters
    ----------
    contact_matrix : 2D array (n, n)
        Whole genome contact matrix.
    partition : 1D array (n_chroms+1,)
        An offset array providing the starting bin of each chromosome and
        whose last element is the last bin of the last chromosome.
    percentile_top : float
        Clip trans blowout pixels above this cutoff.
    percentile_bottom :
        Mask bins with trans coverage below this cutoff.

    Returns
    -------
    2D array (n, n)
        Normalized affinity matrix
    1D array (n,)
        Boolean mask of bins that are not used in the affinity matrix.

    """
    contact_matrix = np.array(contact_matrix)

    # Ensure contact_matrix is square
    if contact_matrix.shape[0] != contact_matrix.shape[1]:
        logger.critical("contact_matrix.shape: %s", contact_matrix.shape)
        raise ValueError("Matrix A is not square.")

    # Adjust partitions if partition[0] != 0
    if partition[0] != 0:
        logger.error("partition[0] != 0.")
        raise ValueError("partition[0] != 0.")

    logger.debug("contact_matrix.shape: %s", contact_matrix.shape)
    n_bins = contact_matrix.shape[0]
    if not (partition[0] == 0 and partition[-1] == n_bins and np.all(np.diff(partition) > 0)):
        logger.warning("partition: %s", partition)
        logger.warning("partition[0]: %s", partition[0])
        logger.warning("partition[-1]: %s", partition[-1])
        logger.warning("np.diff(partition) > 0: %s", np.diff(partition) > 0)
        raise ValueError(f"Not a valid partition. Must be a monotonic sequence from 0 to {n_bins}.")

    # Zero out cis data and create mask for trans
    extents = pairwise(partition)
    part_ids = []
    for n, (i0, i1) in enumerate(extents):
        contact_matrix[i0:i1, i0:i1] = 0
        part_ids.extend([n] * (i1 - i0))
    part_ids = np.array(part_ids)
    is_trans = part_ids[:, None] != part_ids[None, :]

    # Zero out bins nulled out using NaNs
    is_bad_bin = np.nansum(contact_matrix, axis=0) == 0
    contact_matrix[is_bad_bin, :] = 0
    contact_matrix[:, is_bad_bin] = 0

    if np.any(~np.isfinite(contact_matrix)) or np.any(np.isnan(contact_matrix)):
        raise ValueError("Matrix A contains point-wise NaNs or infinite values, not expected for cooler-loaded maps")

    # Filter the heatmap
    is_good_bin = ~is_bad_bin
    is_valid = np.logical_and.outer(is_good_bin, is_good_bin)
    contact_matrix = _filter_heatmap(contact_matrix, is_trans & is_valid, percentile_top, percentile_bottom)
    is_bad_bin = np.nansum(contact_matrix, axis=0) == 0
    contact_matrix[is_bad_bin, :] = 0
    contact_matrix[:, is_bad_bin] = 0

    # Inject decoy cis data, balance and rescale margins to 1
    contact_matrix = _fake_cis(contact_matrix, ~is_trans)
    numutils.set_diag(contact_matrix, 0, 0)
    contact_matrix = numutils.iterative_correction_symmetric(contact_matrix)[0]
    marg = np.r_[np.sum(contact_matrix, axis=0), np.sum(contact_matrix, axis=1)]
    marg = np.mean(marg[marg > 0])
    contact_matrix /= marg

    contact_matrix = _fake_cis(contact_matrix, ~is_trans)
    numutils.set_diag(contact_matrix, 0, 0)
    contact_matrix = numutils.iterative_correction_symmetric(contact_matrix)[0]
    marg = np.r_[np.sum(contact_matrix, axis=0), np.sum(contact_matrix, axis=1)]
    marg = np.mean(marg[marg > 0])
    contact_matrix /= marg

    contact_matrix[is_bad_bin, :] = 0
    contact_matrix[:, is_bad_bin] = 0

    return contact_matrix, is_bad_bin



"""Functions for creating heatmaps of ENCODE ChIP-seq signal."""

from collections.abc import Iterable

import datashader as ds
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from datashader.mpl_ext import dsshow


def plot_components(
    dataset, component_1="PCA1", component_2="PCA2", cluster_col=None, name_col="name", filename=None, remove_ticks=True
):
    """Plot the components of a dataset in a grid.

    Parameters
    ----------
    dataset : pd.DataFrame
        The dataset to plot
    component_1 : str
        The name of the column containing the first component
    component_2 : str
        The name of the column containing the second component
    cluster_col : str
        The name of the column containing the cluster labels
    name_col : str
        The name of the column containing the dataset names
    filename : str
        The filename to save the plot to
    remove_ticks : bool
        Whether to remove the axis ticks

    """
    n_files = len(dataset[name_col].unique())
    grid_size = int(np.ceil(np.sqrt(n_files)))
    _, axs = plt.subplots(grid_size, grid_size, figsize=(12, 12))
    xmin = dataset[component_1].min()
    xmax = dataset[component_1].max()
    ymin = dataset[component_2].min()
    ymax = dataset[component_2].max()
    for i, name in enumerate(dataset[name_col].unique()):
        plotdf = dataset[dataset[name_col] == name]
        if cluster_col is not None:
            plotdf = plotdf.loc[plotdf.good_bin, [component_1, component_2, cluster_col]]
            plotdf[cluster_col] = pd.Categorical(plotdf[cluster_col])
            dsshow(
                plotdf,
                ds.Point(component_1, component_2),
                aggregator=ds.count_cat(cluster_col),
                aspect="auto",
                ax=axs[i // grid_size, i % grid_size],
                x_range=(xmin, xmax),
                y_range=(ymin, ymax),
            )
        else:
            plotdf = plotdf.loc[plotdf.good_bin, [component_1, component_2]]
            dsshow(
                plotdf,
                ds.Point(component_1, component_2),
                aspect="auto",
                ax=axs[i // grid_size, i % grid_size],
                x_range=(xmin, xmax),
                y_range=(ymin, ymax),
            )
        axs[i // grid_size, i % grid_size].set_title(name)

        if remove_ticks:
            axs[i // grid_size, i % grid_size].set_xticks([])
            axs[i // grid_size, i % grid_size].set_yticks([])

    if filename is not None:
        plt.savefig(filename)


def plot_signal(
    dataset,
    component_1: str,
    component_2: str,
    signals: Iterable,
    labels: list | None = None,
    filename=None,
    remove_ticks=True,
):
    """Plot signal values overlaying an embedding.

    Parameters
    ----------
    dataset : pd.DataFrame
        The dataset to plot
    component_1 : str
        The name of the column containing the first component
    component_2 : str
        The name of the column containing the second component
    signals : list
        The list of columns containing the signal values
    labels : list
        The list of columns containing the cluster labels
    filename : str
        The filename to save the plot to
    remove_ticks : bool
        Whether to remove the axis ticks

    """
    df = dataset.dropna().copy()

    # If length of unique label values is >20, set all values >20 to 20
    if labels is not None:
        for label in labels:
            if len(df[label].unique()) > 20:
                df[label] = df[label].apply(lambda x: x if x < 20 else 20)

    n_plots = len(signals) + len(labels)
    grid_size = int(np.ceil(np.sqrt(n_plots)))
    _, axs = plt.subplots(grid_size, grid_size, figsize=(8, 8))
    xmin = df[component_1].min()
    xmax = df[component_1].max()
    ymin = df[component_2].min()
    ymax = df[component_2].max()

    # Plot categorical labels
    if labels is not None:
        for i, label in enumerate(labels):
            df[label] = pd.Categorical(df[label])
            dsshow(
                df,
                ds.Point(component_1, component_2),
                aggregator=ds.count_cat(label),
                aspect="auto",
                ax=axs[i // grid_size, i % grid_size],
                x_range=(xmin, xmax),
                y_range=(ymin, ymax),
            )
            axs[i // grid_size, i % grid_size].set_title(label)
            if remove_ticks:
                axs[i // grid_size, i % grid_size].set_xticks([])
                axs[i // grid_size, i % grid_size].set_yticks([])

    # Plot signal values
    for i, signal in enumerate(signals):
        dsshow(
            df,
            ds.Point(component_1, component_2),
            aggregator=ds.mean(signal),
            aspect="auto",
            ax=axs[(i + len(labels)) // grid_size, (i + len(labels)) % grid_size],
            x_range=(xmin, xmax),
            y_range=(ymin, ymax),
            cmap="seismic",
        )
        axs[(i + len(labels)) // grid_size, (i + len(labels)) % grid_size].set_title(signal)
        if remove_ticks:
            axs[(i + len(labels)) // grid_size, (i + len(labels)) % grid_size].set_xticks([])
            axs[(i + len(labels)) // grid_size, (i + len(labels)) % grid_size].set_yticks([])

    if filename is not None:
        plt.savefig(filename)


def create_dotplot(
    df: pd.DataFrame, cluster_column: str, signal_columns: list[str], threshold: float, filename: str | None = None
):
    """Create a dot plot of signal values for each cluster.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to plot
    cluster_column : str
        The name of the column containing the cluster labels
    signal_columns : List[str]
        The names of the columns containing the signal values
    threshold : float
        The threshold to use for the percentage calculation
    filename : str
        The filename to save the plot to

    """
    # Calculate fractions and means
    fractions_above_threshold = (
        df[[*signal_columns, cluster_column]]
        .groupby(cluster_column)
        .apply(lambda x: (x[signal_columns] > threshold).sum() / len(x))
        .reset_index()
    )
    means = df[[*signal_columns, cluster_column]].groupby(cluster_column).mean().reset_index()

    # Z-scale means by each signal
    for signal in signal_columns:
        means[signal] = (means[signal] - means[signal].mean()) / means[signal].std()

    # Prepare the plot data
    plot_data = []
    for col in signal_columns:
        for cluster in df[cluster_column].unique():
            fraction = fractions_above_threshold.loc[fractions_above_threshold[cluster_column] == cluster, col].values[
                0
            ]
            mean = means.loc[means[cluster_column] == cluster, col].values[0]
            plot_data.append({"Cluster": cluster, "Signal": col, "Fraction": fraction, "Mean": mean})

    # Plotting
    plot_df = pd.DataFrame(plot_data)
    fig, ax = plt.subplots(figsize=(8, 4))
    sns.scatterplot(
        data=plot_df, x="Cluster", y="Signal", size="Fraction", hue="Mean", ax=ax, legend=False, palette="seismic"
    )
    plt.xlabel("Cluster")
    plt.ylabel("Signal")

    if filename:
        plt.savefig(filename)



from unittest.mock import MagicMock

from jointly_hic.parser import JointlyCommandLineInterface


def test_joint_pca_cli_embedding(mocker):
    """Test JointPCACommandLineInterface for embedding."""
    parser = JointlyCommandLineInterface()

    # Test parsing arguments for embedding
    args = parser.parser.parse_args(
        "embed --mcools data.mcool --chrom-limit 5 --resolution 100000 --assembly hg38 --method NMF".split()
    )
    assert args.func.__name__ == "run_embedding"
    assert args.mcools == ["data.mcool"]
    assert args.chrom_limit == 5
    assert args.resolution == 100000
    assert args.assembly == "hg38"
    assert args.log_level == "INFO"
    assert args.method == "NMF"


def test_joint_pca_cli_post(mocker):
    """Test JointPCACommandLineInterface fpr parser."""
    parser = JointlyCommandLineInterface()

    # Test parsing arguments for post-processing
    args = parser.parser.parse_args(
        "post-process --parquet-file data.parquet --output test --umap-neighbours 30 100 500 "
        "--kmeans-clusters 5 10 15 20".split()
    )
    assert args.func.__name__ == "run_post_process"
    assert args.parquet_file == "data.parquet"
    assert args.output == "test"
    assert args.umap_neighbours == [30, 100, 500]
    assert args.kmeans_clusters == [5, 10, 15, 20]
    assert args.log_level == "INFO"

    # Patch PostProcessor
    mocker.patch("jointly_hic.core.post_processor.PostProcessor")
    args.func(args)


def test_joint_pca_cli_trajectory(mocker):
    """Test JointPCACommandLineInterface for trajectory analyzer."""
    parser = JointlyCommandLineInterface()

    # Test parsing arguments for trajectory analysis
    args = parser.parser.parse_args(
        "trajectory --parquet-file data.parquet --output test --umap-neighbours 30 100 500 "
        "--kmeans-clusters 5 10 15 20".split()
    )
    assert args.func.__name__ == "run_trajectory"
    assert args.parquet_file == "data.parquet"
    assert args.output == "test"
    assert args.umap_neighbours == [30, 100, 500]
    assert args.kmeans_clusters == [5, 10, 15, 20]
    assert args.log_level == "INFO"

    # Patch TrajectoryAnalyzer
    mocker.patch("jointly_hic.core.trajectory_analyzer.TrajectoryAnalyzer")
    args.func(args)



"""Utilities for working with ENCODE data."""

import os.path
import shutil
import tempfile
from multiprocessing.pool import ThreadPool

import bbi
import pandas as pd
import requests

# Set the cache directory, default to ~/cache if not set
CACHE_DIR = os.getenv("CACHE_DIR", tempfile.gettempdir())


def pick_bigwig(filelist, output_type="fold change over control") -> str:
    """Find the best bigWig from a list: ENCODE4, max bioreps, fold change over control, etc.

    Parameters
    ----------
    filelist : list
        List of file paths from an ENCODE experiment
    output_type : str
        The output type to filter on, default is "fold change over control"

    Returns
    -------
    str
        The accession of the best bigWig file

    """
    file_candidates = []
    with ThreadPool(50) as pool:
        filedata_results = pool.map(
            lambda url: requests.get(url, timeout=60).json(),
            [f"https://www.encodeproject.org{fpath}?format=json" for fpath in filelist],
        )

    for filedata in filedata_results:
        if filedata["file_format"] != "bigWig":
            continue
        if filedata["assembly"] != "GRCh38":
            continue
        if filedata["status"] != "released":
            continue
        if filedata["award"]["viewing_group"] != "ENCODE4":
            continue
        if filedata["output_type"] != output_type:
            continue

        # Now, we know it's an ENCODE bigWig aligned to hg38
        result = {
            "accession": filedata["title"],
            "reps": filedata["biological_replicates"],
            "replen": len(filedata["biological_replicates"]),
        }
        file_candidates.append(result)
    if len(file_candidates) == 0:
        return None
    files = pd.DataFrame(file_candidates).sort_values("replen", ascending=False).reset_index()
    return str(files["accession"][0])


def get_bigwigs_from_experiment_report(seach_term: str) -> dict:
    """Create a dict of {name: accession} for fold change ENCODE4 bigwigs from a url to an experiment report."""
    url = (
        f"https://www.encodeproject.org/search/?type=Experiment&searchTerm={seach_term}&"
        f"assay_title=Histone+ChIP-seq&status=released&assembly=GRCh38&format=json"
    )
    response = requests.get(url, timeout=60)
    payload = response.json()
    result = {}
    for experiment in payload["@graph"]:  # The list of experiments
        biosample = experiment["biosample_ontology"]["term_name"]
        target = experiment["target"]["label"]
        exp_accession = experiment["accession"]
        filelist = [f["@id"] for f in experiment["files"]]
        bigwig_accession = pick_bigwig(filelist)
        if bigwig_accession is None:
            continue
        acc_dict = {f"{biosample}.{target}.{exp_accession}": bigwig_accession}
        result.update(acc_dict)
    return result


def get_signal(path, bins):
    """Get signal from bigwig file."""
    with bbi.open(path) as bw:
        signal = bw.stackup(bins["chrom"], bins["start"], bins["end"], bins=1).ravel()
    return signal


class EncodeFile:
    """Implement context manager that downloads, caches, and moves an ENCODE file to/from temporary storage."""

    def __init__(self, accession: str, extension: str = "bw"):
        """Initialize the EncodeFile object."""
        self.accession = accession
        self.extension = extension
        self.cache_path = f"{CACHE_DIR}/{self.accession}.{self.extension}"
        self.local_file = f"{self.accession}.{self.extension}"

    def __enter__(self):
        """Download file if not exists, and move to temp."""
        if os.path.exists(self.cache_path):
            shutil.copyfile(self.cache_path, self.local_file)
        else:
            if self.extension == "bw":
                extension = "bigWig"
            else:
                extension = self.extension
            url = f"https://www.encodeproject.org/files/{self.accession}/@@download/{self.accession}.{extension}"
            with requests.get(url, stream=True, allow_redirects=True, timeout=60) as r:
                with open(self.local_file, "wb") as f:
                    shutil.copyfileobj(r.raw, f)
            shutil.copy(self.local_file, self.cache_path)
        return self.local_file

    def __exit__(self, exc_type, exc_value, traceback):
        """Remove the temporary file."""
        try:
            os.remove(self.local_file)
        except FileNotFoundError:
            pass



"""Exntension of sklearn.decomposition._BasePCA to support SVD decomposition."""

from numbers import Integral
from typing import ClassVar

import numpy as np
from scipy import sparse
from sklearn.base import _fit_context
from sklearn.decomposition._base import _BasePCA
from sklearn.utils import check_random_state, gen_batches
from sklearn.utils._arpack import _init_arpack_v0
from sklearn.utils._param_validation import Interval, Real, StrOptions
from sklearn.utils.extmath import randomized_svd, safe_sparse_dot, svd_flip
from sklearn.utils.sparsefuncs import incr_mean_variance_axis


class SparseIncrementalSVD(_BasePCA):
    """Sparse Incremental Singular Value Decomposition (siSVD).

    Linear dimensionality reduction using Singular Value Decomposition of
    the data, keeping only the most significant singular vectors to
    project the data to a lower dimensional space.
    Derived from :class:`sklearn.decomposition.IncrementalPCA` and
    :class:`sklearn.decomposition.TruncatedSVD`.
    Unlike IncrementalPCA:
        * This estimator is designed to work on sparse `csr_matrix` input.
        * This estimator does NOT center the data before applying SVD.
    Unlike TruncatedSVD:
        * This estimator supports incremental (minibatch) learning via the
          `partial_fit` method.

    Parameters
    ----------
    n_components : int, default=None
        Number of components to keep. If ``n_components`` is ``None``,
        then ``n_components`` is set to **one less than**
        ``min(n_samples, n_features)``due to the way the ARPACK solver works.
    batch_size : int, default=None
        The number of samples to use for each batch. Only used when calling
        ``fit``. If ``batch_size`` is ``None``, then ``batch_size``
        is inferred from the data and set to ``5 * n_features``, to provide a
        balance between approximation accuracy and memory consumption.

    See Also
    --------
    sklearn.decomposition.IncrementalPCA
    sklearn.decomposition.TruncatedSVD

    """

    _parameter_constraints: ClassVar[dict] = {
        "n_components": [Interval(Integral, 1, None, closed="left"), None],
        "whiten": ["boolean"],
        "copy": ["boolean"],
        "batch_size": [Interval(Integral, 1, None, closed="left"), None],
        "algorithm": [StrOptions({"arpack", "randomized"})],
        "n_iter": [Interval(Integral, 0, None, closed="left")],
        "n_oversamples": [Interval(Integral, 1, None, closed="left")],
        "power_iteration_normalizer": [StrOptions({"auto", "OR", "LU", "none"})],
        "random_state": ["random_state"],
        "tol": [Interval(Real, 0, None, closed="left")],
    }

    def __init__(
        self,
        n_components=None,
        *,
        batch_size=None,
        algorithm="randomized",
        n_iter=5,
        n_oversamples=10,
        power_iteration_normalizer="auto",
        random_state=None,
        tol=0.0,
        whiten=False,
        copy=True,
    ):
        """Initialize parameters for SparseIncrementalSVD."""
        self.n_components = n_components
        self.whiten = whiten
        self.copy = copy
        self.batch_size = batch_size
        self.algorithm = algorithm
        self.n_iter = n_iter
        self.n_oversamples = n_oversamples
        self.power_iteration_normalizer = power_iteration_normalizer
        self.random_state = random_state
        self.tol = tol

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y=None):
        """Fit the model with X, using minibatches of size batch_size.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : object
            Returns the instance itself.

        """
        self.components_ = None
        self.n_samples_seen_ = 0
        self.mean_ = 0.0
        self.var_ = 0.0
        self.singular_values_ = None
        self.explained_variance_ = None
        self.explained_variance_ratio_ = None

        X = self._validate_data(
            X,
            accept_sparse=["csr", "csc", "lil"],
            copy=self.copy,
            dtype=[np.float64, np.float32],
        )
        n_samples, n_features = X.shape

        if self.batch_size is None:
            self.batch_size_ = 5 * n_features
        else:
            self.batch_size_ = self.batch_size

        for batch in gen_batches(n_samples, self.batch_size_, min_batch_size=self.n_components or 0):
            X_batch = X[batch]
            self.partial_fit(X_batch)

        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def partial_fit(self, X, y=None):
        """Incremental fit with X. All of X is processed as a single batch.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y : Ignored
            Not used, present for API consistency by convention.
        check_input : bool, default=True
            Run check_array on X.

        Returns
        -------
        self : object
            Returns the instance itself.

        """
        random_state = check_random_state(self.random_state)
        if not hasattr(self, "components_"):
            self.components_ = None
        first_pass = self.components_ is None
        n_samples, n_features = X.shape

        if self.n_components is None:
            if self.components_ is None:
                self.n_components_ = min(n_samples, n_features) - 1
            else:
                self.n_components_ = self.components_.shape[0]
        elif not self.n_components <= n_features:
            raise ValueError(
                "n_components=%r invalid for n_features=%d, need more rows than columns for IncrementalPCA processing"
            )
        elif not self.n_components <= n_samples:
            raise ValueError("n_components=%r must be less or equal to the batch number of samples %d.")
        else:
            self.n_components_ = self.n_components

        if (self.components_ is not None) and (self.components_.shape[0] != self.n_components_):
            raise ValueError(
                "Number of input features has changed from %i "
                "to %i between calls to partial_fit! Try "
                "setting n_components to a fixed value."
            )

        # Convert X to sparse matrix, if not already
        X = sparse.csr_matrix(X)

        # This is the first partial_fit
        if first_pass:
            self.n_samples_seen_ = 0
            self.mean_ = np.zeros(X.shape[1])
            self.var_ = np.zeros(X.shape[1])

        col_mean, col_var, n_total_samples = incr_mean_variance_axis(
            X,
            axis=0,
            last_mean=self.mean_,
            last_var=self.var_,
            last_n=np.repeat(self.n_samples_seen_, X.shape[1]),
        )
        n_total_samples = n_total_samples[0]

        if self.n_samples_seen_ > 0:
            X = sparse.vstack(
                (
                    # The first `n_components` rows are the scaled singular vectors
                    # Each vector is length `n_features` (the previous basis)
                    self.singular_values_.reshape((-1, 1)) * self.components_,
                    # The new batch of data (n_samples, n_features) make up the remaining rows
                    X,
                    # Because we don't mean-center, we don't append a mean correction row
                )
            )

        if self.algorithm == "arpack":
            v0 = _init_arpack_v0(min(X.shape), random_state)
            U, S, Vt = sparse.linalg.svds(X, k=self.n_components_, tol=self.tol, v0=v0)
            S = S[::-1]
            U, Vt = svd_flip(U[:, ::-1], Vt[::-1])

        elif self.algorithm == "randomized":
            if self.n_components > X.shape[1]:
                raise ValueError(f"n_components({self.n_components}) must be <= n_features({X.shape[1]}).")
            U, S, Vt = randomized_svd(
                X,
                self.n_components,
                n_iter=self.n_iter,
                n_oversamples=self.n_oversamples,
                power_iteration_normalizer=self.power_iteration_normalizer,
                random_state=random_state,
            )

        explained_variance = S**2 / (n_total_samples - 1)
        explained_variance_ratio = S**2 / np.sum(col_var * n_total_samples)
        self.n_samples_seen_ = n_total_samples
        self.mean_ = col_mean
        self.var_ = col_var

        self.components_ = Vt
        self.singular_values_ = S
        self.explained_variance_ = explained_variance
        self.explained_variance_ratio_ = explained_variance_ratio
        return self

    def transform(self, X):
        """Apply dimensionality reduction to X.

        X is projected on the first principal components previously extracted
        from a training set, using minibatches of size batch_size if X is
        sparse.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        Returns
        -------
        X_new : ndarray of shape (n_samples, n_components)
            Projection of X in the first principal components.

        Examples
        --------
        >>> import numpy as np
        >>> from sklearn.decomposition import IncrementalPCA
        >>> X = np.array([[-1, -1], [-2, -1], [-3, -2],
        ...               [1, 1], [2, 1], [3, 2]])
        >>> ipca = IncrementalPCA(n_components=2, batch_size=3)
        >>> ipca.fit(X)
        IncrementalPCA(batch_size=3, n_components=2)
        >>> ipca.transform(X) # doctest: +SKIP

        """
        X_transformed = safe_sparse_dot(X, self.components_.T)
        return X_transformed



import os
import tempfile
from unittest.mock import patch

import numpy as np
import pytest
import pandas as pd
import h5py
import yaml

from jointly_hic.hdf5db.hdf5db import (
    create_hdf5,
    load_bins_from_embeddings,
    store_bins,
    load_bins,
    store_experiment_metadata,
    load_experiment_metadata,
    store_embeddings,
    load_embeddings,
    add_bigwig_track,
    add_encode_tracks_to_hdf5,
    load_track_metadata,
    load_tracks,
    JointDb,
)

from jointly_hic.hdf5db.experiments2yaml import main as experiments2yaml_main

EMBEDDINGS_FILE = "test-data.pq"
TEST_EMBEDDINGS = os.path.join(os.path.dirname(__file__), EMBEDDINGS_FILE)
ACCESSION_COLUMN = "hic_accession"


@pytest.fixture(scope="function")
def test_file():
    with tempfile.NamedTemporaryFile(suffix=".yaml") as temp_file:
        temp_file_path = temp_file.name
        yield temp_file_path

    assert not os.path.exists(temp_file_path)


@pytest.fixture
def test_data():
    return pd.read_parquet(TEST_EMBEDDINGS)


@pytest.fixture
def test_metadata_yaml():
    # Define the temporary YAML file path
    with tempfile.NamedTemporaryFile(suffix=".yaml") as temp_file:
        temp_file_path = temp_file.name

        # Call the main function from the experiments2yaml script to generate the YAML file
        metadata_columns = ["biosample", "hic_accession", "filename"]  # Define your metadata columns

        # Generate the YAML file
        experiments2yaml_main(
            parquet_file=TEST_EMBEDDINGS,
            accession_column=ACCESSION_COLUMN,
            metadata_columns=metadata_columns,
            yaml_file=temp_file_path,
        )

        yield temp_file_path

    assert not os.path.exists(temp_file_path)


def test_create_hdf5(test_file):
    # Create an empty HDF5 file
    create_hdf5(test_file)
    assert os.path.exists(test_file)

    # Open the file and check if the groups exist
    with h5py.File(test_file, "r") as hdf:
        assert "bins" in hdf
        assert "embeddings" in hdf
        assert "metadata" in hdf
        assert "tracks" in hdf


def test_load_bins_from_embeddings():
    # Load bins from the test data
    bins = load_bins_from_embeddings(TEST_EMBEDDINGS)

    # Check that the dataframe is not empty and contains the expected columns
    assert not bins.empty
    assert "chrom" in bins.columns
    assert "start" in bins.columns
    assert "end" in bins.columns
    assert "good_bin" in bins.columns


def test_bin_loading(test_file, test_data):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store bins information into the HDF5 file
    store_bins(test_file, TEST_EMBEDDINGS)

    # Open the HDF5 file and check if the bins datasets exist
    with h5py.File(test_file, "r") as hdf:
        bins_group = hdf["bins"]
        assert "chrom" in bins_group
        assert "start" in bins_group
        assert "end" in bins_group
        assert "good_bin" in bins_group
        assert "bin_name" in bins_group

    # Load the bins from the HDF5 file
    bins_df = load_bins(test_file)

    # Check that the dataframe is not empty and contains the expected columns
    assert not bins_df.empty
    assert "chrom" in bins_df.columns
    assert "start" in bins_df.columns
    assert "end" in bins_df.columns
    assert "good_bin" in bins_df.columns
    assert "bin_name" in bins_df.columns

    # Load originals, sort by chrom, start, end, and verify columns chrom, start, end, good_bin
    original_df = test_data[["chrom", "start", "end", "good_bin"]]
    original_df = original_df.drop_duplicates().sort_values(by=["chrom", "start", "end"])
    original_df = original_df[["chrom", "start", "end", "good_bin"]]
    original_df = original_df.reset_index(drop=True)
    bins_df = bins_df.sort_values(by=["chrom", "start", "end"])
    bins_df = bins_df[["chrom", "start", "end", "good_bin"]]
    bins_df = bins_df.reset_index(drop=True)
    pd.testing.assert_frame_equal(original_df, bins_df)


def test_experiment_metadata(test_file, test_metadata_yaml):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store experiment metadata
    store_experiment_metadata(test_file, test_metadata_yaml)

    # Open the HDF5 file and check if the experiment metadata is stored
    with h5py.File(test_file, "r") as hdf:
        metadata_group = hdf["metadata"]
        assert "experiment_metadata" in metadata_group

    # Load the experiment metadata
    metadata_df = load_experiment_metadata(test_file)

    # Check that the dataframe is not empty and contains the expected columns
    assert not metadata_df.empty
    assert "accession" in metadata_df.columns
    assert "biosample" in metadata_df.columns

    # Load original data from YAML for comparison
    with open(test_metadata_yaml) as file:
        original_metadata_list = yaml.safe_load(file)
    original_metadata_df = pd.DataFrame(original_metadata_list).astype(str)

    # Compare the loaded data with the original data
    pd.testing.assert_frame_equal(original_metadata_df, metadata_df)


def test_store_embeddings(test_file, test_metadata_yaml, test_data):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store experiment metadata
    store_experiment_metadata(test_file, test_metadata_yaml)

    # Store bins
    store_bins(test_file, TEST_EMBEDDINGS)

    # Store embeddings
    store_embeddings(test_file, TEST_EMBEDDINGS, ACCESSION_COLUMN)

    # Open the HDF5 file and check if the embeddings datasets exist
    with h5py.File(test_file, "r") as hdf:
        embeddings_group = hdf["embeddings"]
        # Verify if the embeddings datasets are created
        for column in test_data.columns:
            if (
                "PCA" in column
                or "NMF" in column
                or "SVD" in column
                or "umap" in column
                or "leiden" in column
                or "kmeans" in column
            ):
                assert column in embeddings_group


def test_load_embeddings(test_file, test_metadata_yaml, test_data):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store experiment metadata
    store_experiment_metadata(test_file, test_metadata_yaml)

    # Store bins
    store_bins(test_file, TEST_EMBEDDINGS)

    # Store embeddings
    store_embeddings(test_file, TEST_EMBEDDINGS, ACCESSION_COLUMN)

    # Load experiment metadata and bins
    experiment_metadata = load_experiment_metadata(test_file)
    bins = load_bins(test_file)

    # Define embeddings to load
    embeddings_to_load = ["PCA1", "PCA2"]

    # Load the embeddings
    embeddings_df = load_embeddings(test_file, embeddings_to_load, experiment_metadata, bins)

    # Check that the dataframe is not empty and contains the expected columns
    assert not embeddings_df.empty
    assert "accession" in embeddings_df.columns
    assert "bin_name" in embeddings_df.columns
    assert "PCA1" in embeddings_df.columns
    assert "PCA2" in embeddings_df.columns

    # Verify the content of the loaded embeddings matches the stored data
    original_data = test_data[["hic_accession", "chrom", "start", "end", *embeddings_to_load]].rename(
        columns={"hic_accession": "accession"}
    )
    original_data["bin_name"] = [
        f"{chrom}:{start}-{end}"
        for chrom, start, end in zip(original_data["chrom"], original_data["start"], original_data["end"])
    ]
    original_data = original_data.sort_values(["accession", "bin_name"]).set_index(["accession", "bin_name"])
    original_data = original_data[embeddings_to_load]
    print(original_data)
    loaded_data = embeddings_df.sort_values(["accession", "bin_name"]).set_index(["accession", "bin_name"])[
        embeddings_to_load
    ]
    print(loaded_data)

    pd.testing.assert_frame_equal(original_data, loaded_data)

    # Test JointDb access
    db = JointDb(test_file)
    embeddings = db.get_embeddings(embeddings_to_load)
    pd.testing.assert_frame_equal(embeddings, embeddings_df)


class MockEncodeFile:
    """Mock class to simulate the opening of a file."""

    def __init__(self, accession, filetype):
        self.accession = accession
        self.filetype = filetype

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def mock_get_signal(bigwig, bins_df):
    return np.random.rand(len(bins_df))


@pytest.fixture
def mock_track_metadata():
    data = [
        {"accession": "track1", "assay": "H3K27ac", "experiment": "exp1", "biosample": "cell1"},
        {"accession": "track2", "assay": "H3K4me3", "experiment": "exp2", "biosample": "cell2"},
    ]
    return pd.DataFrame(data)


def test_bigwig_tracks(test_file, test_data):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store bins
    store_bins(test_file, TEST_EMBEDDINGS)
    bins = load_bins(test_file)

    # Mock signal data
    signal = pd.DataFrame({"signal": np.random.rand(bins.shape[0])})
    accession = "track1"
    assay = "H3K27ac"
    experiment = "exp1"
    biosample = "cell1"

    # Add bigwig track
    add_bigwig_track(test_file, signal, accession, assay, experiment, biosample)

    # Check if the track and metadata are stored correctly
    with h5py.File(test_file, "r") as hdf:
        tracks_group = hdf["tracks"]
        assert accession in tracks_group
        loaded_data = np.array(tracks_group[accession][:], dtype=float).reshape(-1)
        expected_data = np.array(signal["signal"].values, dtype=float).reshape(-1)
        assert np.allclose(loaded_data, expected_data)

        metadata_group = hdf["metadata/track_metadata"]
        track_group = metadata_group[accession]
        assert track_group.attrs["assay"] == assay
        assert track_group.attrs["experiment"] == experiment
        assert track_group.attrs["biosample"] == biosample

    # Load the track signal data
    bins_df = load_bins(test_file)
    track_signals = load_tracks(test_file, [accession], bins_df, drop_bad_bins=False)
    loaded_signal = track_signals[accession].values

    # Assert signal matches loaded track_signal column
    assert np.allclose(loaded_signal, expected_data)


@patch("jointly_hic.hdf5db.hdf5db.EncodeFile", new=MockEncodeFile)
@patch("jointly_hic.hdf5db.hdf5db.get_signal", new=mock_get_signal)
def test_add_encode_tracks_to_hdf5(test_file, mock_track_metadata):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store bins (mock data)
    store_bins(test_file, TEST_EMBEDDINGS)

    # Add mock tracks and their metadata
    add_encode_tracks_to_hdf5(mock_track_metadata, test_file)

    # Verify that the tracks and their metadata are correctly added
    with h5py.File(test_file, "r") as hdf:
        tracks_group = hdf["tracks"]
        metadata_group = hdf["metadata/track_metadata"]

        for track in mock_track_metadata.itertuples():
            assert track.accession in tracks_group
            track_group = metadata_group[track.accession]
            assert track_group.attrs["assay"] == track.assay
            assert track_group.attrs["experiment"] == track.experiment
            assert track_group.attrs["biosample"] == track.biosample

    # Test JointDb accession
    db = JointDb(test_file)
    track_metadata = db.track_metadata
    assert track_metadata["accession"].tolist() == mock_track_metadata["accession"].tolist()


@patch("jointly_hic.hdf5db.hdf5db.EncodeFile", new=MockEncodeFile)
@patch("jointly_hic.hdf5db.hdf5db.get_signal", new=mock_get_signal)
def test_load_track_metadata(test_file, mock_track_metadata):
    # Create an empty HDF5 file
    create_hdf5(test_file)

    # Store bins (mock data)
    store_bins(test_file, TEST_EMBEDDINGS)

    # Add mock tracks and their metadata
    add_encode_tracks_to_hdf5(mock_track_metadata, test_file)

    # Load the track metadata
    loaded_metadata = load_track_metadata(test_file)

    # Check that the metadata matches the original mock data
    pd.testing.assert_frame_equal(
        loaded_metadata.sort_values(by="accession").reset_index(drop=True),
        mock_track_metadata.sort_values(by="accession").reset_index(drop=True),
    )


def test_jointdb_initialization(test_file):
    # Create an empty HDF5 file
    create_hdf5(test_file)
    joint_db = JointDb(test_file)
    assert joint_db.hdf5_file == test_file


def test_jointdb_bins(test_file, test_data):
    # Create an empty HDF5 file and store bins
    create_hdf5(test_file)
    store_bins(test_file, TEST_EMBEDDINGS)

    # Initialize JointDb and access bins
    joint_db = JointDb(test_file)
    bins_df = joint_db.bins

    # Check if the bins data matches the test data
    original_bins = (
        test_data[["chrom", "start", "end", "good_bin"]].drop_duplicates().sort_values(by=["chrom", "start", "end"])
    )
    original_bins.reset_index(drop=True, inplace=True)
    bins_df = bins_df.sort_values(by=["chrom", "start", "end"]).reset_index(drop=True)[
        ["chrom", "start", "end", "good_bin"]
    ]
    pd.testing.assert_frame_equal(original_bins, bins_df)


def test_jointdb_experiment_metadata(test_file, test_metadata_yaml):
    # Create an empty HDF5 file and store experiment metadata
    create_hdf5(test_file)
    store_experiment_metadata(test_file, test_metadata_yaml)

    # Initialize JointDb and access experiment metadata
    joint_db = JointDb(test_file)
    experiment_metadata_df = joint_db.experiment_metadata

    # Load original data from YAML for comparison
    with open(test_metadata_yaml) as file:
        original_metadata_list = yaml.safe_load(file)
    original_metadata_df = pd.DataFrame(original_metadata_list).astype(str)

    # Compare the loaded data with the original data
    pd.testing.assert_frame_equal(original_metadata_df, experiment_metadata_df)


def test_jointdb_get_embeddings(test_file, test_metadata_yaml, test_data):
    # Create an empty HDF5 file, store metadata, bins, and embeddings
    create_hdf5(test_file)
    store_experiment_metadata(test_file, test_metadata_yaml)
    store_bins(test_file, TEST_EMBEDDINGS)
    store_embeddings(test_file, TEST_EMBEDDINGS, ACCESSION_COLUMN)

    # Initialize JointDb and get embeddings
    joint_db = JointDb(test_file)
    embeddings_to_load = ["PCA1", "PCA2"]
    embeddings_df = joint_db.get_embeddings(embeddings_to_load)

    # Check that the dataframe is not empty and contains the expected columns
    assert not embeddings_df.empty
    assert "accession" in embeddings_df.columns
    assert "bin_name" in embeddings_df.columns
    assert "PCA1" in embeddings_df.columns
    assert "PCA2" in embeddings_df.columns

    # Verify the content of the loaded embeddings matches the stored data
    original_data = test_data[["hic_accession", "chrom", "start", "end", *embeddings_to_load]].rename(
        columns={"hic_accession": "accession"}
    )
    original_data["bin_name"] = [
        f"{chrom}:{start}-{end}"
        for chrom, start, end in zip(original_data["chrom"], original_data["start"], original_data["end"])
    ]
    original_data = original_data.sort_values(["accession", "bin_name"]).set_index(["accession", "bin_name"])
    original_data = original_data[embeddings_to_load]
    loaded_data = embeddings_df.sort_values(["accession", "bin_name"]).set_index(["accession", "bin_name"])[
        embeddings_to_load
    ]

    pd.testing.assert_frame_equal(original_data, loaded_data)



"""Main command class for Joint PCA Hi-C."""

import gzip
import logging
import os
import pickle
from datetime import datetime

import bioframe
import numpy as np
import pandas as pd
from cooler import Cooler
from pandas import DataFrame
from sklearn.decomposition import IncrementalPCA, MiniBatchNMF

from jointly_hic.core.config import JointlyConfiguration, PostProcessingConfig, TrajectoryAnalysisConfig
from jointly_hic.methods.multivec import make_multivec, nanmean_agg
from jointly_hic.methods.sparse_incremental_svd import SparseIncrementalSVD
from jointly_hic.methods.trans_cooler_utils import normalized_affinity_matrix_from_trans

logger = logging.getLogger("joint_pca")


class JointlyDecomposer:
    """Main command class for Joint PCA Hi-C.

    JointlyDecomposer performs incremental PCA or NMF docomposition of multiple input 3D chromatin contact frequency
    matrices from mcool files. JointlyDecomposer can be run from the command line or imported as a module. It takes a
    JointlyConfig object as input and returns a tuple of the output embeddings and a PostProcessingConfig object,
    for use configuring the post-processing step. Use JointlyDecomposer.run() to run the decomposition.

    The fields in the configuration are:
        mcools: List of input mcool files
        output: Output file prefix
        resolution: Resolution of input mcool files
        assembly: Genome assembly of input mcool files
        components: Number of components to compute
        chrom_limit: Number of chromosomes to use
        method: Method to use for decomposition. Either 'PCA' or 'NMF'
        percentile_top: Top percentile of contact frequencies to threshold at
        percentile_bottom: Bottom percentile of contact frequencies to remove
        batch_size: Batch size for incremental decomposition
        log_level: Log level for logging

    Returns
    -------
        Tuple of output embeddings and PostProcessingConfig object

    """

    def __init__(self, configuration: JointlyConfiguration):
        """Initialize JointlyDecomposer."""
        self.configuration = configuration
        self.union_bad_bins: np.ndarray | None = None
        self._partition = None
        logger.info("JointlyDecomposer :: configuration: \n\n%s", configuration)

        # Initialize model
        if configuration.method == "PCA":
            self.model = IncrementalPCA(n_components=configuration.components)
        elif configuration.method == "NMF":
            self.model = MiniBatchNMF(n_components=configuration.components)
        elif configuration.method == "SVD":
            self.model = SparseIncrementalSVD(n_components=configuration.components)
        else:
            raise ValueError(f"Unknown method: {configuration.method}")

        # Get chromosome sizes
        self.chromosome_sizes = self.get_chromosome_sizes()

    def run(self) -> tuple[DataFrame, PostProcessingConfig, TrajectoryAnalysisConfig]:
        """Run joint PCA."""
        # Get start time
        start_time = datetime.now()

        # Get union set of bad bins
        self.set_union_bad_bins()

        # Iterate through files and compute low dimension embeddings
        for filename in self.configuration.mcools:
            # Decompose cooler file
            self.decompose_cooler_file(
                filename=filename,
            )
        logger.info("Model training complete. Training time: %s.", str(datetime.now() - start_time))

        # Compute embeddings using fully trained model
        logger.info("Computing embeddings using fully trained model...")
        output_embeddings = self.compute_output_embeddings().reset_index(drop=True)

        # Output file names
        output_prefix = (
            f"{self.configuration.output}_{self.configuration.method}-{self.configuration.components}_"
            f"{self.configuration.resolution}bp_{self.configuration.assembly}"
        )
        parquet_filename = f"{output_prefix}_embeddings.pq"
        csv_filename = f"{output_prefix}_embeddings.csv.gz"
        model_filename = f"{output_prefix}_model.pkl.gz"

        # Save results
        logger.info("Saving results...")
        self.save_model(model_filename)
        output_embeddings.to_csv(csv_filename, index=False)
        output_embeddings.to_parquet(parquet_filename, index=False)
        logger.debug(f"Output embeddings for filenames:\n{output_embeddings.filename.value_counts()}")
        logger.info("Saved embeddings to '%s' and '%s'.", csv_filename, parquet_filename)

        # Create configuration for post-processing
        post_config = PostProcessingConfig(
            parquet_file=parquet_filename,
            output=output_prefix,
            method=self.configuration.method,
            log_level=self.configuration.log_level,
        )
        trajectory_config = TrajectoryAnalysisConfig(
            parquet_file=parquet_filename,
            output=output_prefix,
            method=self.configuration.method,
            log_level=self.configuration.log_level,
        )

        # Log total time
        logger.info("Finished joint PCA in %s.", str(datetime.now() - start_time))
        return output_embeddings.copy(), post_config, trajectory_config

    @property
    def partition(self) -> np.ndarray:
        """Get partition numbers for each chromosome."""
        if self._partition is not None:
            return self._partition

        partition = None
        for cooler_file in self.configuration.mcools:
            clr = Cooler(cooler_file + f"::/resolutions/{self.configuration.resolution}")
            partition_sizes = [clr.offset(chrom) for chrom in self.chromosome_sizes.index]
            if partition is None:
                partition = np.r_[
                    partition_sizes,
                    clr.extent(self.chromosome_sizes.index[-1])[1],
                ]
            else:
                if not np.all(partition == np.r_[partition_sizes, clr.extent(self.chromosome_sizes.index[-1])[1]]):
                    raise ValueError("Chromosome partitions do not match across input files")
            logger.debug("Loaded partitions: [%s]", ", ".join([str(x) for x in partition]))
            self._partition = partition
        return self._partition

    def get_chromosome_sizes(self) -> pd.Series:
        """Download chromosome sizes using bioframe."""
        chromosome_sizes = bioframe.fetch_chromsizes(self.configuration.assembly.lower())
        # chromosome_sizes is a pandas Series with chromosome names as index and chromosome sizes as values
        # Filter to only the first chrom_limit chromosomes
        chromosome_sizes = chromosome_sizes.iloc[: self.configuration.chrom_limit]
        logger.info("Loaded chromosome sizes for specified assembly: %s", self.configuration.assembly)
        logger.info("Chromosome sizes:\n%s", chromosome_sizes)
        return chromosome_sizes

    def set_union_bad_bins(self) -> None:
        """Compute union of bad bins (where weight is NaN) across all input files.

        Returns
        -------
            Array of bad bins (where weight is NaN) across all input files

        """
        logger.info("Beginning to compute union set of NaN bins..")
        start_time = datetime.now()
        bad_bins: np.ndarray | None = None
        bins: pd.DataFrame | None = None

        # Bins are bad when the weight is NaN
        for filename in self.configuration.mcools:
            clr = Cooler(f"{filename}::/resolutions/{self.configuration.resolution}")
            lo = self.partition[0]
            hi = self.partition[-1]
            bins = clr.bins()[lo:hi]
            logger.debug("Loaded bins from cooler file '%s' with shape '%s'.", filename, bins.shape)
            if bad_bins is None:
                bad_bins = np.isnan(bins["weight"].to_numpy())
            else:
                bad_bins = np.logical_or(bad_bins, np.isnan(bins["weight"].to_numpy()))
            logger.debug("Percent bad bins: %s.", 100 * np.sum(bad_bins) / bad_bins.shape[0])

        # Load excluded_bins from exclusion list if provided
        if self.configuration.exclusion_list is not None:
            logger.info("Loading exclusion list from: %s", self.configuration.exclusion_list)
            excluded_bins = pd.read_csv(
                self.configuration.exclusion_list, sep="\t", header=None, names=["chrom", "start", "end"]
            )
            # Exclude bins overlapping an excluded_bin
            overlap = bioframe.overlap(bins, excluded_bins, how="inner", return_index=True, return_input=False)
            ix = overlap["index"].to_numpy(dtype="int64")
            bad_bins[ix] = True
            logger.debug("Percent bad bins: %s.", 100 * np.sum(bad_bins) / bad_bins.shape[0])

        # Preprocess all input files and include the bad bins from normalization
        assert isinstance(bad_bins, np.ndarray)
        self.union_bad_bins = bad_bins
        for filename in self.configuration.mcools:
            self.preprocess_matrix(filename, ignore_cache=True)
            logger.debug("Percent bad bins: %s.", 100 * np.sum(bad_bins) / bad_bins.shape[0])

        logger.info("Loaded union set of bad bins in: %s.", str(datetime.now() - start_time))
        logger.info(
            "Percent of bins that are bad: %s.", 100 * np.sum(self.union_bad_bins) / self.union_bad_bins.shape[0]
        )

        # Save bins to file
        logger.debug("Shape of bad_bin array: %s.", self.union_bad_bins.shape)
        bins["bad_bin"] = self.union_bad_bins
        logger.debug("Shape of bins dataframe: %s.", bins.shape)
        logger.debug("Saving bins to file...")
        bins.to_csv(f"{self.configuration.output}_bins.csv.gz", index=False)

    def preprocess_matrix(self, filename: str, ignore_cache: bool = False) -> np.ndarray:
        """Load matrix from cooler and preprocess it.

        Loads matrix from cooler file, computes cis-masked, balanced, affinity matrix, and removes bad bins.
        The result is cached to disk for future use.

        Args:
        ----
            filename: Input .mcool filename containing contact frequency matrix
            ignore_cache: If True, ignore cached matrix and recompute

        Returns:
        -------
            np.ndarray: Preprocessed contact frequency matrix

        """
        # Load matrix from cooler file
        start_time = datetime.now()
        clr = Cooler(filename + f"::/resolutions/{self.configuration.resolution}")
        partition = self.partition
        lo = partition[0]
        hi = partition[-1]

        # Load from disk if file exists
        matrix_filename = f"{filename.removesuffix('.mcool')}_preprocessed_{lo}_{hi}.npy"
        if os.path.exists(matrix_filename) and not ignore_cache:
            logger.info("Loading preprocessed matrix from disk...")
            matrix = np.load(matrix_filename)
            logger.debug(
                "Loaded preprocessed matrix with shape '%s' from disk in %s.",
                matrix.shape,
                str(datetime.now() - start_time),
            )

            # Remove bad bins
            bin_logical = np.logical_not(self.union_bad_bins)
            matrix = matrix[bin_logical, :]
            matrix = matrix[:, bin_logical]
            logger.debug("Removed bad bins from matrix. New shape: %s", matrix.shape)
            return matrix

        # Load contact frequency matrix
        matrix = clr.matrix(balance=True)[lo:hi, lo:hi]
        logger.debug(
            "Loaded contact frequency matrix with shape '%s' in %s.", matrix.shape, str(datetime.now() - start_time)
        )

        # Compute cis-masked, balanced, affinity matrix
        start_time = datetime.now()
        # Adjust partition so that it starts at 0
        partition = partition - partition[0]
        matrix, normalization_bad_bins = normalized_affinity_matrix_from_trans(
            matrix,
            partition,
            self.configuration.percentile_top,
            self.configuration.percentile_bottom,
        )

        # Save matrix to disk for future use
        np.save(matrix_filename, matrix)

        # Update self.union_bad_bins with normalization_bad_bins
        self.union_bad_bins = np.logical_or(self.union_bad_bins, normalization_bad_bins)
        assert isinstance(self.union_bad_bins, np.ndarray)

        logger.debug(
            "Computed cis-masked, balanced, affinity matrix of shape '%s' in %s.",
            matrix.shape,
            str(datetime.now() - start_time),
        )

        # Remove bad bins
        bin_logical = np.logical_not(self.union_bad_bins)
        matrix = matrix[bin_logical, :]
        matrix = matrix[:, bin_logical]
        logger.debug("Removed bad bins from matrix. New shape: %s", matrix.shape)
        return matrix

    def minibatch_fit(self, matrix: np.ndarray):
        """Fit model using partial fit and minibatching input matrix.

        Args:
        ----
            matrix: Input matrix
            batch_size: Batch size

        """
        num_bins = matrix.shape[0]
        num_batches = int(np.ceil(num_bins / self.configuration.batch_size))
        batch_indices = np.array_split(np.arange(num_bins), num_batches)
        logger.debug("Split matrix into %s batches.", num_batches)

        # Iterate over batches
        for i, batch_index in enumerate(batch_indices):
            start_time = datetime.now()
            batch = matrix[batch_index, :]

            # Fit batch
            logger.debug("Fitting batch: %s with nrows: %s.", i + 1, batch.shape[0])
            self.model.partial_fit(batch)
            logger.debug("Finished fitting batch: %s in %s seconds.", i + 1, str(datetime.now() - start_time))

    def decompose_cooler_file(self, filename):
        """Decompose a single cooler file."""
        logger.info("Computing dimensionality reduction for input file: '%s'...", filename)
        start_time = datetime.now()
        matrix = self.preprocess_matrix(filename)
        logger.debug("Finished preprocessing matrix for '%s' in %s.", filename, str(datetime.now() - start_time))

        # Compute incremental fit
        self.minibatch_fit(
            matrix,
        )
        logger.info(
            "Finished decomposition for '%s' in %s.",
            filename,
            str(datetime.now() - start_time),
        )

    def save_model(self, filename):
        """Save model to disk using pickle."""
        start_time = datetime.now()
        with gzip.open(filename, "wb") as f:
            pickle.dump(self.model, f)
        logger.info("Saved model to '%s' in %s.", filename, str(datetime.now() - start_time))

    def compute_output_embeddings(self) -> pd.DataFrame:
        """Iterate through files and compute embeddings."""
        output = []
        for filename in self.configuration.mcools:
            output.append(self.compute_output_embeddings_single_file(filename))
        return pd.concat(output)

    def compute_output_embeddings_single_file(self, filename) -> pd.DataFrame:
        """Compute embeddings for a single input using the trained model."""
        logger.info("Computing embeddings for input file: '%s'...", filename)
        lo = self.partition[0]
        hi = self.partition[-1]

        # Load matrix from cooler file
        matrix = self.preprocess_matrix(filename)

        # Compute embeddings
        embeddings = self.model.transform(matrix)

        # Move embeddings into array of original shape
        full_embeddings = np.zeros((hi - lo, self.configuration.components), dtype=np.float64)
        full_embeddings[:] = np.nan
        full_embeddings[np.logical_not(self.union_bad_bins)] = embeddings

        # Add bin names and convert to dataframe
        bins = Cooler(f"{filename}::/resolutions/{self.configuration.resolution}").bins()[lo:hi]
        result = pd.DataFrame(
            full_embeddings,
            columns=[f"{self.configuration.method}{i + 1}" for i in range(self.configuration.components)],
            dtype=np.float64,
        )
        bins = bins.reset_index(drop=True)
        result = result.reset_index(drop=True)
        good_bin_df = pd.DataFrame({"good_bin": np.logical_not(self.union_bad_bins)}).reset_index(drop=True)
        filename_df = pd.DataFrame({"filename": [filename] * result.shape[0]}).reset_index(drop=True)
        result = pd.concat(
            [
                bins,
                good_bin_df,
                filename_df,
                result,
            ],
            axis=1,
        )

        # Make multivec output
        make_multivec(
            outpath=f"{filename.removesuffix('.mcool')}_{self.configuration.method}.mv5",
            df=result,
            feature_names=[f"{self.configuration.method}{i + 1}" for i in range(min(self.configuration.components, 8))],
            base_res=self.configuration.resolution,
            chromsizes=self.chromosome_sizes.to_dict(),
            tilesize=1024,
            agg=nanmean_agg,
            chunksize=int(1e6),
        )
        return result



"""Script to produce a YAML file of experiment metadata from a parquet file."""

import argparse
import logging

import pandas as pd
import yaml

# Set up logging
logger = logging.getLogger("joint_pca")


def verify_metadata_consistency(df: pd.DataFrame, accession_column: str, metadata_columns: list[str]) -> list[dict]:
    """Verify that each metadata value is the same for all rows with the same accession.

    Args:
    ----
        df (pd.DataFrame): DataFrame containing the data.
        accession_column (str): Column name used as the accession identifier.
        metadata_columns (List[str]): List of other metadata columns to include.

    Returns:
    -------
        List[dict]: List of dictionaries with consistent metadata per accession.

    """
    metadata_list = []

    for accession, group in df.groupby(accession_column):
        metadata_dict = {"accession": accession}

        for column in metadata_columns:
            unique_values = group[column].unique()
            if len(unique_values) > 1:
                raise ValueError(f"Inconsistent values found in column '{column}' for accession '{accession}'.")
            metadata_dict[column] = unique_values[0]

        metadata_list.append(metadata_dict)

    return metadata_list


def write_metadata_to_yaml(metadata_list: list[dict], yaml_file: str) -> None:
    """Write metadata to a YAML file.

    Args:
    ----
        metadata_list (List[dict]): List of dictionaries containing metadata.
        yaml_file (str): Path to the output YAML file.

    """
    with open(yaml_file, "w") as file:
        yaml.dump(metadata_list, file, default_flow_style=False)


def main(parquet_file: str, accession_column: str, metadata_columns: list[str], yaml_file: str) -> None:
    """Load parquet data, verify metadata, and write to YAML.

    Args:
    ----
        parquet_file (str): Path to the parquet file containing data.
        accession_column (str): Column name used as the accession identifier.
        metadata_columns (List[str]): List of other metadata columns to include.
        yaml_file (str): Path to the output YAML file.

    """
    df = pd.read_parquet(parquet_file)
    metadata_list = verify_metadata_consistency(df, accession_column, metadata_columns)
    write_metadata_to_yaml(metadata_list, yaml_file)
    logger.info(f"Metadata successfully written to {yaml_file}")


def run_main(args):
    """Run the main function with arguments."""
    main(args.parquet_file, args.accession_column, args.metadata_columns, args.yaml_file)


def configure_embedding_metadata_parser(embedding_metadata_parser):
    """Configure the parser for the embedding2yaml script."""
    embedding_metadata_parser.add_argument(
        "--parquet-file", required=True, help="Path to the parquet file containing data."
    )
    embedding_metadata_parser.add_argument("--accession-column", required=True, help="Name of the accession column.")
    embedding_metadata_parser.add_argument(
        "--metadata-columns", nargs="+", required=True, help="List of other metadata columns."
    )
    embedding_metadata_parser.add_argument("--yaml-file", required=True, help="Path to the output YAML file.")
    embedding_metadata_parser.set_defaults(func=run_main)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Produce a YAML file of experiment metadata.")
    configure_embedding_metadata_parser(parser)
    arguments = parser.parse_args()
    run_main(arguments)
